#ifndef AJLIB_DEF_H
#define AJLIB_DEF_H

#define NOMINMAX
#include <windows.h>
#define _DO_NOT_DECLARE_INTERLOCKED_INTRINSICS_IN_MEMORY
#include <intrin.h>
#include <string>
#include <sstream>
#include <iomanip>

typedef unsigned char Byte;
typedef unsigned short Word;
typedef unsigned int DWord;
typedef unsigned long long int QWord;
typedef signed char Int3;
typedef signed short int Int4;
typedef signed int Int5;
typedef signed long long int Int6;

#define CODE_MACRO(x) do { x } while (false)

#define IF_TRUE_THROW(expr,exception) CODE_MACRO( \
    if (expr) \
        throw exception; \
)

#define IF_FALSE_THROW(expr) IF_TRUE_THROW(!(expr), Exception())
#define IF_MINUS_ONE_THROW(expr) IF_TRUE_THROW((expr) == -1, Exception())
#define IF_ZERO_THROW(expr) IF_TRUE_THROW((expr) == 0, Exception())
#define IF_NULL_THROW(expr) IF_TRUE_THROW((expr) == NULL, Exception())
#define IF_NONZERO_THROW(expr) IF_TRUE_THROW((expr) != 0, Exception())

#define IF_ERROR_THROW(expr) CODE_MACRO( \
    HRESULT hrMacro = (expr); \
    IF_TRUE_THROW(FAILED(hrMacro), Exception(hrMacro)); \
)

#define IF_ZERO_CHECK_THROW_LAST_ERROR(expr) CODE_MACRO( \
    if (expr == 0) \
        IF_FALSE_THROW(GetLastError() == 0); \
)


#define BEGIN_CHECKED \
    try { \
        try

#define END_CHECKED \
        catch(std::bad_alloc&) { \
            throw Exception(E_OUTOFMEMORY); \
        } \
        catch(std::exception&) { \
            throw Exception(E_FAIL); \
        } \
    } \
    catch


void alert(std::wstring message, HWND hWnd = NULL)
{
    MessageBox(hWnd, message.c_str(), L"Error", MB_OK | MB_ICONERROR);
}

void assertTrue(bool success, std::wstring message=L"", HWND hWnd = NULL)
{
    if (!success) {
        std::wstring output(L"Assertion failed");
        if (message.length() > 0)
            output += L": " + message;
        alert(output, hWnd);
        _CrtDbgBreak();
    }
}


class Exception;


class Uncopyable
{
public:
    Uncopyable() { }
private:
    Uncopyable(const Uncopyable& other);
    const Uncopyable& operator=(const Uncopyable& other);
};


template<class T> class ComPtr
{
public:
    ComPtr() : _p(NULL) { }
    ComPtr(const ComPtr<T>& spT) : _p(spT._p) { _p->AddRef(); }
    ComPtr(T* pT) : _p(pT) { _p->AddRef(); }
    const ComPtr<T>& operator=(const ComPtr<T>& spT)
    {
        _p = spT._p;
        _p->AddRef();
        return *this;
    }
    template<class U> ComPtr(const ComPtr<U>& spU,
        const IID* piid = &__uuidof(T))
    {
        IF_FALSE_THROW(spU.valid());
        IF_ERROR_THROW(
            spU->QueryInterface(*piid, reinterpret_cast<void**>(&_p)));
    }
    ~ComPtr() { reset(); }
	T** operator&()
	{
        if (valid())
		    assertTrue(false, L"Re-initializing ComPtr");
		return &_p;
	}
	T* operator->() const
    {
        if (!valid())
            assertTrue(false, L"Using uninitialized ComPtr");
        return _p;
    }
    bool valid() const { return _p != NULL; }
    void reset()
    {
		if (valid())
			_p->Release();
        _p = NULL;
    }
    operator T*() { return _p; }
    operator const T*() const { return _p; }
private:
	T* _p;
};


class Bstr
{
public:
    Bstr() : m_bstr(NULL) { }
    ~Bstr() { SysFreeString(m_bstr); }
    operator std::wstring() const { return std::wstring(m_bstr); }
    BSTR* operator&()
    {
        assertTrue(m_bstr==NULL, L"Re-initializing Bstr");
        return &m_bstr;
    }
private:
    BSTR m_bstr;
};


class LocalString
{
public:
    LocalString() : m_str(NULL) { }
    ~LocalString() { LocalFree(m_str); }
    operator std::wstring() const { return std::wstring(m_str); }
    LPWSTR* operator&()
    {
        assertTrue(m_str==NULL, L"Re-initializing LocalString");
        return &m_str;
    }
private:
    LPWSTR m_str;
};


class Exception
{
public:
	Exception(HRESULT hr)
	{
        ComPtr<IErrorInfo> spIEI;
        if (GetErrorInfo(0, &spIEI) == S_OK) {
            Bstr bstrDescription;
            spIEI->GetDescription(&bstrDescription);
            _description = bstrDescription;
        }
        else
            setFromErrorCode(hr);
	}
    Exception()
    {
        DWORD dwLastError = GetLastError();
        if (dwLastError == 0) {
            // If there was really no error we wouldn't be here. Avoid emitting
            // messages like "Error: Success"
            dwLastError = E_FAIL;
        }
        setFromErrorCode(dwLastError);
    }
    Exception(std::wstring description) { _description = description; }
	void display(HWND hWnd = NULL) { alert(_description, hWnd); }
private:
    void setFromErrorCode(DWORD dwErrorCode)
    {
        LocalString strMessage;
        DWORD dwFormatted = FormatMessage(
            FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM |
                FORMAT_MESSAGE_IGNORE_INSERTS,
            NULL,
            dwErrorCode,
            0,
            reinterpret_cast<LPWSTR>(&strMessage),
            0,
            NULL);
        if (dwFormatted == 0) {
            std::wstringstream s;
            s << L"FormatMessage failed: 0x";
            s << std::hex;
            s << std::setfill(L'0');
            s << std::setw(8);
            s << GetLastError();
            _description = s.str();
            return;
        }
        _description = strMessage;
    }
protected:
    std::wstring _description;
};


class ComInitializer : Uncopyable
{
public:
    ComInitializer() { IF_ERROR_THROW(CoInitialize(NULL)); }
    ~ComInitializer() { CoUninitialize(); }
};


class Event : Uncopyable
{
public:
    Event()
    {
        m_hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
        IF_NULL_THROW(m_hEvent);
    }
    ~Event() { CloseHandle(m_hEvent); }
    operator HANDLE() const { return m_hEvent; }
    void set() { IF_ZERO_THROW(SetEvent(m_hEvent)); }
    void reset() { IF_ZERO_THROW(ResetEvent(m_hEvent)); }
    void wait() { IF_FALSE_THROW(WaitForSingleObject(m_hEvent, INFINITE)==0); }
private:
    HANDLE m_hEvent;
};


class Thread : Uncopyable
{
public:
    Thread()
      : _started(false)
    {
        _hThread = CreateThread(
            NULL, 0, threadStaticProc, this, CREATE_SUSPENDED, NULL);
        IF_NULL_THROW(_hThread);
    }
    ~Thread() { CloseHandle(_hThread); }
    void setPriority(int nPriority)
    {
        IF_ZERO_THROW(SetThreadPriority(_hThread, nPriority));
    }
    void join()
    {
        if (_started)
            IF_FALSE_THROW(
                WaitForSingleObject(_hThread, INFINITE)==WAIT_OBJECT_0);
    }
    void start() { IF_MINUS_ONE_THROW(ResumeThread(_hThread)); }

private:
    static DWORD WINAPI threadStaticProc(LPVOID lpParameter)
    {
        Thread* thread = reinterpret_cast<Thread*>(lpParameter);
        thread->_started = true;
        thread->threadProc();
        return 0;
    }

    const Thread& operator=(const Thread& other);
    Thread(const Thread& other);

    virtual void threadProc() { }

    HANDLE _hThread;
    bool _started;
};


class Mutex : Uncopyable
{
public:
    Mutex() { InitializeCriticalSection(&_cs); }
    ~Mutex() { DeleteCriticalSection(&_cs); }
    void lock() { EnterCriticalSection(&_cs); }
    void unlock() { LeaveCriticalSection(&_cs); }
    bool tryLock() { return TryEnterCriticalSection(&_cs) != 0; }
private:
    CRITICAL_SECTION _cs;
};


class Lock : Uncopyable
{
public:
    Lock() : _mutex(0) { }
    Lock(Mutex* mutex) : _mutex(mutex) { _mutex->lock(); }

    ~Lock()
    {
        if (_mutex)
            _mutex->unlock();
    }

    bool tryAcquire(Mutex* mutex)
    {
        if (mutex->tryLock()) {
            _mutex = mutex;
            return true;
        }
        return false;
    }

private:
    Mutex* _mutex;
};


int pumpMessages()
{
    BOOL bRet;
    do {
        MSG msg;
        bRet = GetMessage(&msg, NULL, 0, 0);
        IF_MINUS_ONE_THROW(bRet);
        if (bRet != 0) {
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        }
        else
            return static_cast<int>(msg.wParam);
    } while (bRet != 0);
    return 0;
}


class IdleProcessor
{
public:
    virtual void idle() = 0;
};


int pumpMessages(IdleProcessor* idle)
{
    BOOL fMessage;
    MSG msg;
    do {
        do {
            fMessage = PeekMessage(&msg, NULL, 0U, 0U, PM_REMOVE);
            if (!fMessage)
                break;
            if (msg.message == WM_QUIT)
                break;
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        } while (true);
        if (msg.message == WM_QUIT)
            break;
        idle->idle();
    } while (true);
    return static_cast<int>(msg.wParam);
}


template<class T> T max(const T& a, const T& b)
{ 
    if (a>b)
        return a;
    return b;
}


template<class T> T min(const T& a, const T& b)
{
    if (a<b)
        return a;
    return b;
}


template<class T> T max(const T& a, const T& b, const T& c)
{
    return max(a,max(b,c));
}


template<class T> T min(const T& a, const T& b, const T& c)
{
    return min(a,min(b,c));
}


template<class T> T max(const T& a, const T& b, const T& c, const T& d)
{
    return max(a,max(b,c,d));
}


template<class T> T min(const T& a, const T& b, const T& c, const T& d)
{
    return min(a,min(b,c,d));
}


// Finds the smallest n such that 1<<n >= v
int log2(DWord v)
{
    if (v==0)
        throw Exception(std::wstring(
            L"Singularity error: attempted to find logarithm of zero"));
    int n=0;
    if ((v&0xffff0000)!=0) { n+=16; v>>=16; }
    if ((v&0xff00)!=0) { n+=8; v>>=8; }
    if ((v&0xf0)!=0) { n+=4; v>>=4; }
    if ((v&0xc)!=0) { n+=2; v>>=2; }
    if ((v&2)!=0) { n+=1; v>>=1; }
    return n;

    
    //__asm {
    //    bsr eax, v
    //    mov v, eax
    //}
    //return v;
}


// A reference counted pointer to an instance of T
template<class T> class ReferenceCountedPointer
{
public:
    ReferenceCountedPointer(const T& t)
      : _body(new Body(t))
    { }
    ReferenceCountedPointer()
      : _body(0)
    { }
    ReferenceCountedPointer(const ReferenceCountedPointer& x)
      : _body(x._body)
    {
        ++_body->_referenceCount;
    }
    ~ReferenceCountedPointer() { release(); }
    const ReferenceCountedPointer& operator=(const ReferenceCountedPointer& x)
    {
        release();
        _body = x._body;
        ++_body->_referenceCount;
        return *this;
    }
    const ReferenceCountedPointer& operator=(const T& t)
    {
        release();
        _body = new Body(t);
        return *this;
    }
    T& operator*() const { return _body->_t; }
    T* operator->() const { return &_body->_t; }
    bool valid() const { return _body!=0; }
private:
    void release()
    {
        if (_body != 0) {
            --_body->_referenceCount;
            if (_body->_referenceCount==0)
                delete _body;
        }                                                                                                                                                                                                   
        _body = 0;
    }

    struct Body
    {
        Body(const T& t)
          : _t(t),
            _referenceCount(0)
        { }
        T _t;
        int _referenceCount;
    };
    Body* _body;
};


#endif // AJLIB_DEF_H