#ifndef AJLIB_DIRECTX_H
#define AJLIB_DIRECTX_H

#include <d3d9.h>
#include <dxerr9.h>
#include <math.h>
#include "linkedlist.h"
#include "colour.h"
#include "user.h"

// This is the exception that we throw if any DirectX methods fail.
class DirectXException : public Exception
{
public:
    DirectXException(HRESULT hr)
    {
        _description = DXGetErrorDescription9(hr);
    }
};

#define IF_ERROR_THROW_DX(expr) CODE_MACRO( \
    HRESULT hrMacro = (expr); \
    IF_TRUE_THROW(FAILED(hrMacro), DirectXException(hrMacro)); \
)

// This class encapsulates an IDirect3DTexture9 - i.e. one texture.
class Direct3DTexture
{
public:
    Direct3DTexture() : _locked(false) { }

    void create(IDirect3DDevice9* device, Vector2<int> size)
    {
        _device = device;
        IF_ERROR_THROW_DX(device->CreateTexture(
            size.x,           // Width
            size.y,           // Height
            1,                // Levels
            0,                // Usage
            D3DFMT_X8R8G8B8,  // Format
            D3DPOOL_MANAGED,  // Pool
            &_texture,        // ppTexture
            NULL));           // pSharedHandle
    }

    void lock()
    {
        IF_ERROR_THROW_DX(_texture->LockRect(
            0,           // Level
            &_lockData,  // pLockedRect
            NULL,        // pRect
            0));         // Flags
        _locked = true;
    }

    void unlock() { IF_ERROR_THROW_DX(doUnlock()); }

    ~Direct3DTexture()
    {
        if (_locked)
            doUnlock();
    }

    void setTexture()
    {
        IF_ERROR_THROW_DX(_device->SetTexture(
            0,           // Sampler
            _texture));  // pTexture
    }

    Byte* data() const { return reinterpret_cast<Byte*>(_lockData.pBits); }

    int pitch() const { return _lockData.Pitch; }
private:
    HRESULT doUnlock()
    {
        _locked = false;  // Don't reattempt on destruction if call fails
        return _texture->UnlockRect(0);
    }

    ComPtr<IDirect3DTexture9> _texture;
    bool _locked;
    D3DLOCKED_RECT _lockData;
    IDirect3DDevice9* _device;
};

// This class encapsulates an IDirect3DVertexBuffer9 - i.e. the geometry for 
// one object.
class Direct3DVertices
{
public:
    Direct3DVertices() : _locked(false) { }

    void create(IDirect3DDevice9* device)
    {
        _device = device;
        IF_ERROR_THROW_DX(device->CreateVertexBuffer(
            sizeof(Vertex) * 4,           // Length
            0,                            // Usage
            D3DFVF_XYZRHW | D3DFVF_TEX1,  // FVF
            D3DPOOL_MANAGED,              // Pool
            &_vertexBuffer,               // ppVertexBuffer
            NULL));                       // pSharedHandle
        lock();

        Vertex* vertices = data();
        setVertex(0, Vector2<float>(0, 0));
        vertices[0].z = 0.0f;
        vertices[0].rhw = 1.0f;
        vertices[0].u = 0.0f;
        vertices[0].v = 0.0f;

        setVertex(1, Vector2<float>(256, 0));
        vertices[1].z = 0.0f;
        vertices[1].rhw = 1.0f;
        vertices[1].u = 1.0f;
        vertices[1].v = 0.0f;

        setVertex(2, Vector2<float>(256, 256));
        vertices[2].z = 0.0f;
        vertices[2].rhw = 1.0f;
        vertices[2].u = 1.0f;
        vertices[2].v = 1.0f;

        setVertex(3, Vector2<float>(0, 256));
        vertices[3].z = 0.0f;
        vertices[3].rhw = 1.0f;
        vertices[3].u = 0.0f;
        vertices[3].v = 1.0f;
    }

    void lock()
    {
        IF_ERROR_THROW_DX(_vertexBuffer->Lock(
            0,       // OffsetToLock
            0,       // SizeToLock
            &_data,  // ppbData
            0));     // Flags
        _locked = true;
    }

    void unlock() { IF_ERROR_THROW_DX(doUnlock()); }

    ~Direct3DVertices()
    {
        if (_locked)
            doUnlock();
    }

    void setVertex(int n, Vector2<float> position)
    {
        Vertex* vertices = data();
        vertices[n].x = position.x - 0.5f;
        vertices[n].y = position.y - 0.5f;
    }

    void setStreamSource()
    {
        IF_ERROR_THROW_DX(_device->SetStreamSource(
            0,                 // StreamNumber
            _vertexBuffer,     // pStreamData
            0,                 // OffsetInBytes
            sizeof(Vertex)));  // Stride
    }

private:
    struct Vertex
    {
        float x, y, z;
        float rhw;
        float u, v;
    };

    Vertex* data() const { return reinterpret_cast<Vertex*>(_data); }

    HRESULT doUnlock()
    {
        _locked = false;  // Don't reattempt on destruction if call fails
        return _vertexBuffer->Unlock();
    }

    ComPtr<IDirect3DVertexBuffer9> _vertexBuffer;
    bool _locked;
    VOID* _data;
    IDirect3DDevice9* _device;
};

// This class encapsulates a Direct3D adapter
class Direct3DAdapter
{
public:
    Direct3DAdapter() : _adapter(D3DADAPTER_DEFAULT) { }
    explicit Direct3DAdapter(UINT adapter) : _adapter(adapter) { }
    operator UINT() const { return _adapter; }
private:
    UINT _adapter;
};

// This class encapsulates a D3DFORMAT
class Direct3DFormat
{
public:
    operator D3DFORMAT() { return _format; }
private:
    Direct3DFormat(D3DFORMAT format) : _format(format) { }
    friend class Direct3DDisplayMode;
    D3DFORMAT _format;
};

// This class encapsulates a D3DDISPLAYMODE
class Direct3DDisplayMode
{
public:
    int getWidth() const { return _displayMode.Width; }
    int getHeight() const { return _displayMode.Height; }
    int getRefreshRate() const { return _displayMode.RefreshRate; }
    Direct3DFormat getFormat() const
    {
        return Direct3DFormat(_displayMode.Format);
    }
private:
    Direct3DDisplayMode() { }
    D3DDISPLAYMODE* getD3DDisplayModePointer() { return &_displayMode; }
    friend class Direct3D;
    D3DDISPLAYMODE _displayMode;
};

// This class encapsulates a D3DPRESENT_PARAMETERS
class Direct3DPresentationParameters
{
public:
    Direct3DPresentationParameters()
    {
        ZeroMemory(&_presentationParameters, sizeof(_presentationParameters));
        _presentationParameters.BackBufferCount = 1;
        _presentationParameters.MultiSampleType = D3DMULTISAMPLE_NONE;
        _presentationParameters.MultiSampleQuality = 0;
        _presentationParameters.SwapEffect = D3DSWAPEFFECT_DISCARD;
        _presentationParameters.Windowed = TRUE;
        _presentationParameters.EnableAutoDepthStencil = FALSE;
        _presentationParameters.AutoDepthStencilFormat = D3DFMT_D16;
        _presentationParameters.Flags = 0;
        _presentationParameters.FullScreen_RefreshRateInHz = 0;
        _presentationParameters.PresentationInterval =
            D3DPRESENT_INTERVAL_IMMEDIATE;
    }
    void setSize(Vector2<int> size)
    {
        _presentationParameters.BackBufferWidth = size.x;
        _presentationParameters.BackBufferHeight = size.y;
    }

    void setFormat(Direct3DFormat format)
    {
        _presentationParameters.BackBufferFormat = format;
    }

    void setWindow(const Window& window)
    {
        _presentationParameters.hDeviceWindow = window;
    }

    D3DPRESENT_PARAMETERS* operator&() { return &_presentationParameters; }
private:
    D3DPRESENT_PARAMETERS _presentationParameters;
};

// This class encapsulates an IDirect3D9 - i.e. the per-process Direct3D 
// object.
class Direct3D
{
public:
    Direct3D()
    {
        _direct3D = Direct3DCreate9(D3D_SDK_VERSION);
        IF_FALSE_THROW(_direct3D.valid());
    }
    Direct3DDisplayMode getAdapterDisplayMode(
        Direct3DAdapter adapter = Direct3DAdapter()) const
    {
        Direct3DDisplayMode mode;
        IF_ERROR_THROW_DX(_direct3D->GetAdapterDisplayMode(
            adapter, mode.getD3DDisplayModePointer()));
        return mode;
    }
private:
    ComPtr<IDirect3D9> _direct3D;
    template<class Base, class ImageType> friend class Direct3DWindow;

    ComPtr<IDirect3DDevice9> createDevice(
        Direct3DAdapter adapter, D3DDEVTYPE deviceType, const Window& window,
        DWORD behaviorFlags, Direct3DPresentationParameters& parameters) const
    {
        ComPtr<IDirect3DDevice9> device;
        IF_ERROR_THROW_DX(_direct3D->CreateDevice(
            adapter,        // Adapter
            deviceType,     // DeviceType
            window,         // hFocusWindow
            behaviorFlags,  // BehaviorFlags
            &parameters,    // pPresentationParameters
            &device));      // ppReturnedDeviceInterface
        return device;
    }
};

// This class represents an image window that uses Direct3D for painting its
// image. It also encapsulates an IDirect3DDevice9.
template<class Base, class ImageType> class Direct3DWindow : public Base
{
public:
    class Params
    {
        friend class Direct3DWindow;
    public:
        Params(typename Base::Params bp, ImageType* image,
            const Direct3D& direct3D)
          : _bp(bp),
            _image(image),
            _direct3D(&direct3D)
        { }
    private:
        typename Base::Params _bp;
        ImageType* _image;
        const Direct3D* _direct3D;
    };

    Direct3DWindow(Params p) 
      : Base(p._bp), 
        _image(p._image),
        _direct3D(p._direct3D),
        _resizing(false)
    {
        Vector2<int> size = getSize();

        _parameters.setSize(size);
        _parameters.setFormat(_direct3D->getAdapterDisplayMode().getFormat());
        _parameters.setWindow(*this);
        _device = _direct3D->createDevice(
            Direct3DAdapter(), 
            D3DDEVTYPE_HAL, 
            *this, 
            D3DCREATE_SOFTWARE_VERTEXPROCESSING /*| D3DCREATE_FPU_PRESERVE*/
            | D3DCREATE_MULTITHREADED,
            _parameters);
        _image->setDevice(_device);
        setRenderState();
    }

    IDirect3DDevice9* getDevice() { return _device; }

    void invalidate() { IF_ZERO_THROW(InvalidateRect(_hWnd, NULL, FALSE)); }
private:
    Direct3DPresentationParameters _parameters;
    const Direct3D* _direct3D;
    ComPtr<IDirect3DDevice9> _device;        

protected:
    ImageType* _image;

    virtual void draw() { }

    virtual bool onPaint()
    {
        return doPaint();
    }

    virtual bool doPaint()
    {
        PaintHandle paint(*this);
        if (paint.zeroArea())
            return false;
        IF_ERROR_THROW_DX(_device->BeginScene());
        _image->paint();
        IF_ERROR_THROW_DX(_device->EndScene());
        HRESULT hr = _device->Present(
            NULL,    // pSourceRect
            NULL,    // pDestRect
            NULL,    // hDestWindowOverride
            NULL);   // pDirtyRegion
        if (hr == D3DERR_DEVICELOST) {
            IF_ERROR_THROW_DX(_device->Reset(&_parameters));
            setRenderState();
        }
        else
            IF_ERROR_THROW_DX(hr);

        return true;
    }

    virtual LRESULT handleMessage(UINT uMsg, WPARAM wParam, LPARAM lParam)
    {
        switch (uMsg) {
            case WM_PAINT:
                onPaint();
                return 0;

            case WM_SIZE:
                onResize(Vector2<int>(
                    GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam)));
                return 0;

            case WM_EXITSIZEMOVE:
                if (_resizing)
                    doneResize();
                _resizing = false;
                return 0;
        }

        return Base::handleMessage(uMsg, wParam, lParam);
    }

protected:    
    void onResize(Vector2<int> size)
    {
        if (!size.zeroArea()) {
            _parameters.setSize(size);
            IF_ERROR_THROW_DX(_device->Reset(&_parameters));
            setRenderState();

            _image->resize(size);
            invalidate();
            _resizing = true;
        }
    }

private:
    void setRenderState()
    {
        IF_ERROR_THROW_DX(_device->SetVertexShader(NULL));
        IF_ERROR_THROW_DX(_device->SetFVF(D3DFVF_XYZRHW | D3DFVF_TEX1));
        IF_ERROR_THROW_DX(_device->SetRenderState(D3DRS_LIGHTING, FALSE));
        IF_ERROR_THROW_DX(_device->SetSamplerState(
            0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR));
        IF_ERROR_THROW_DX(_device->SetSamplerState(
            0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR));
        IF_ERROR_THROW_DX(_device->SetSamplerState(
            0, D3DSAMP_ADDRESSU, D3DTADDRESS_CLAMP));
        IF_ERROR_THROW_DX(_device->SetSamplerState(
            0, D3DSAMP_ADDRESSV, D3DTADDRESS_CLAMP));
        IF_ERROR_THROW_DX(_device->SetRenderState(D3DRS_SRGBWRITEENABLE, TRUE));
    }
    void doneResize()
    {
        _image->doneResize();
    }

    bool _resizing;
};

// This class represents a square textured tile - i.e. a single Direct3DTexture
// and a matching Direct3DVertices.
class Direct3DTexturedTile : public LinkedListMember<Direct3DTexturedTile>
{
public:
    void setDevice(IDirect3DDevice9* device)
    {
        _device = device;
        _texture.create(
            device, Vector2<int>(texelsPerTile(), texelsPerTile()));
        _vertices.create(device);
        plotBlock(Vector2<int>(0,0), texelsPerTile(), Colour::black());
    }

    void paint()
    {
        _vertices.unlock();
        _vertices.setStreamSource();
        _texture.setTexture();
        IF_ERROR_THROW_DX(_device->DrawPrimitive(
            D3DPT_TRIANGLEFAN,  // PrimitiveType
            0,                  // StartVertex
            2));                // PrimitiveCount
        _vertices.lock();
    }

    void setVertexPosition(int n, Vector2<float> position)
    {
        _vertices.setVertex(n, position);
    }

    void plotBlock(Vector2<int> texel, int size, Colour colour)
    {
        _texture.lock();
        DWord d = colour.asDWord();
        int pitch = _texture.pitch();
        Byte* p = _texture.data() + texel.y*pitch + texel.x*4;
        for (int y=0;y<size;++y) {
            Byte* l = p;
            for (int x=0;x<size;++x) {
                *reinterpret_cast<DWord*>(l) = d;
                l+=4;
            }
            p += pitch;
        }
        _texture.unlock();
    }

    Colour get4TexelAverage(Vector2<int> topLeft)
    {
        _texture.lock();
        Colour c = average(_texture.data() + topLeft.y*_texture.pitch() + 
            topLeft.x*4, _texture.pitch());
        _texture.unlock();
        return c;
    }

    void doubleFrom(Direct3DTexturedTile* lower, Vector2<int> quadrant)
    {
        _texture.lock();
        int pitch = _texture.pitch();
        int lowerPitch = lower->_texture.pitch();
        Byte* sl = lower->_texture.data() + quadrant.x*(texelsPerTile()>>1)*4 +
            quadrant.y*lowerPitch*(texelsPerTile()>>1);
        Byte* dl = _texture.data();
        for (int y=0;y<texelsPerTile();y+=2) {
            DWord* s = reinterpret_cast<DWord*>(sl);
            DWord* d = reinterpret_cast<DWord*>(dl);
            for (int x=0;x<texelsPerTile();x+=2) {
                DWord c = *(s++);
                *(d++) = c;
                *(d++) = c;
            }
            s = reinterpret_cast<DWord*>(sl);
            dl += pitch;
            d = reinterpret_cast<DWord*>(dl);
            for (int x=0;x<texelsPerTile();x+=2) {
                DWord c = *(s++);
                *(d++) = c;
                *(d++) = c;
            }
            dl += pitch;
            sl += lowerPitch;
        }
        _texture.unlock();
    }

    // Shrink down an upper tile into a quadrant of this one using linear
    // interpolation. Used when creating a mip-mapped render tile from a plot
    // tile.
    void halveFrom(Direct3DTexturedTile* upper, Vector2<int> quadrant)
    {
        _texture.lock();
        int pitch = _texture.pitch();
        int upperPitch = upper->_texture.pitch();
        Byte* sl = upper->_texture.data();
        Byte* dl = _texture.data() + quadrant.x*(texelsPerTile()>>1)*4 +
            quadrant.y*pitch*(texelsPerTile()>>1);
        for (int y=0;y<texelsPerTile();y+=2) {
            Byte* s = sl;
            Byte* d = dl;
            for (int x=0;x<texelsPerTile();x+=2) {
                *reinterpret_cast<DWord*>(d) =
                    average(s, upperPitch).asDWord();
                d+=4;
                s+=8;
            }
            sl += upperPitch+upperPitch;
            dl += pitch;
        }
        _texture.unlock();
    }

    // Shrink down an upper tile into a quadrant of this one by taking every
    // 4th texel. Used when zooming out to avoid creating excess detail which
    // looks wrong when you zoom back in again.
    void coarseHalveFrom(Direct3DTexturedTile* upper, Vector2<int> quadrant)
    {
        _texture.lock();
        int pitch = _texture.pitch();
        int upperPitch = (upper->_texture.pitch())<<1;
        Byte* sl = upper->_texture.data();
        Byte* dl = _texture.data() + quadrant.x*(texelsPerTile()>>1)*4 +
            quadrant.y*pitch*(texelsPerTile()>>1);
        for (int y=0;y<texelsPerTile();y+=2) {
            Byte* s = sl;
            Byte* d = dl;
            for (int x=0;x<texelsPerTile();x+=2) {
                *reinterpret_cast<DWord*>(d) = *reinterpret_cast<DWord*>(s);
                d+=4;
                s+=8;
            }
            sl += upperPitch;
            dl += pitch;
        }
        _texture.unlock();
    }

    static int texelsPerTileLevel()
    {
        // 256*256 works well with most 3D graphics devices.
        return 8;
    }

private:
    static Colour average(Byte* p, int pitch)
    {
        int d0 = *reinterpret_cast<DWord*>(p);
        int d1 = *reinterpret_cast<DWord*>(p + 4);
        int d2 = *reinterpret_cast<DWord*>(p + pitch);
        int d3 = *reinterpret_cast<DWord*>(p + pitch + 4);
        UFix8p8 r = ((d0&0xff) + (d1&0xff) + (d2&0xff) + (d3&0xff))>>2;
        UFix8p8 g = (((d0>>8)&0xff) + ((d1>>8)&0xff) + ((d2>>8)&0xff) + 
            ((d3>>8)&0xff))>>2;
        UFix8p8 b = (((d0>>16)&0xff) + ((d1>>16)&0xff) + ((d2>>16)&0xff) + 
            ((d3>>16)&0xff))>>2;
        return Colour(r, g, b);
    }

    static int texelsPerTile() { return 1<<texelsPerTileLevel(); }

    IDirect3DDevice9* _device;

protected:
    Direct3DTexture _texture;
    Direct3DVertices _vertices;
};

#endif // AJLIB_DIRECTX_H