#include "def.h"
#include "complex.h"
#include "fix.h"
#include "colour.h"
#include "directx.h"
#include "fractal.h"
#include "linkedlist.h"
#include <vector>
#include <list>
#include <algorithm>
#include <map>

// Forward declarations
class FractalProcessor;
template<class FractalProcessor> class Direct3DMipMapTileGrid;
template<class FractalProcessor> class Matrix;
template<class FractalProcessor> class WorkQueueList;
template<class FractalProcessor> class Screen;

// Typedefs for commonly used template instantiations
typedef Vector2<int> Vector;
typedef Fixed<16, Int5, Int6> Fix16p16;

class MColour : public Colour
{
public:
    MColour() { }
    MColour(unsigned int iterations)
    {
        if (iterations == 0) {
            _r = 0;
            _g = 0;
            _b = 0;
        }
        else {
            static float f = 1024/log(256.0f);
            int i = static_cast<int>(log(static_cast<float>(iterations))*f);
            _r = gammaSinewave[i&1023];
            _g = gammaSinewave[(i+341)&1023];
            _b = gammaSinewave[(i+683)&1023];
        }
    }
};

// This class represents an infinitely pannable image with a fixed resolution.
// It is composed of a rectangular grid of Direct3DTexturedTile objects.
template<class FractalProcessor> class Direct3DTileGrid
{
    typedef Direct3DMipMapTileGrid<FractalProcessor> Direct3DMipMapTileGrid;
public:
    Direct3DTileGrid(Direct3DMipMapTileGrid* mipMapTileGrid)
      : _gridSize(1, 1),
        _gridTiles(1, 0),
        _mipMapTileGrid(mipMapTileGrid)
    { }

    void paint()
    {
        int i=0;
        for (Direct3DTexturedTile* tile = _activeTiles.getNext(); tile != 0;
            tile = _activeTiles.getNext(tile))
            tile->paint();
    }

    ~Direct3DTileGrid() { removeActiveTiles(); }

    void moveGrid(Vector newGridSize, Vector newTopLeft)
    {
        if (newGridSize == _gridSize && newTopLeft == Vector(0, 0))
            return;
        std::vector<Direct3DTexturedTile*> newGrid(newGridSize.x*newGridSize.y,
            0);
        for (int y=0;y<_gridSize.y;++y)
            for (int x=0;x<_gridSize.x;++x) {
                Vector oldPos(x, y);
                Direct3DTexturedTile* tile = tileAt(oldPos);
                if (tile != 0) {
                    Vector newPos = oldPos - newTopLeft;
                    if (newPos.inside(newGridSize))
                        newGrid[newPos.y*newGridSize.x+newPos.x] = tile;
                    else
                        _mipMapTileGrid->moveToReserve(tile);
                }
            }
        _gridSize = newGridSize;
        _gridTiles = newGrid;
    }

    void updateVertices(int g=0)
    {
        for (int y=0; y<_gridSize.y; ++y)
            for (int x=0; x<_gridSize.x; ++x) {
                Vector tilePosition(x, y);
                Direct3DTexturedTile* tile = tileAt(tilePosition);
                if (tile == 0)
                    continue;
                Vector t(tilePosition<<texelsPerTileLevel());
                tile->setVertexPosition(0, _mipMapTileGrid->pixelFromTexel(t));
                t.x += texelsPerTile();
                tile->setVertexPosition(1, _mipMapTileGrid->pixelFromTexel(t));
                t.y += texelsPerTile();
                tile->setVertexPosition(2, _mipMapTileGrid->pixelFromTexel(t));
                t.x -= texelsPerTile();
                tile->setVertexPosition(3, _mipMapTileGrid->pixelFromTexel(t));
            }
    }

    // Set the pixels within the square whose top-left corner is "topLeft" and
    // whose edges are "size" texels to colour "colour". This square is assumed
    // to be entirely within a single tile.
    void plotBlock(Vector topLeft, int size, Colour colour)
    {
        tileAt(topLeft>>texelsPerTileLevel())->plotBlock(
            topLeft&(texelsPerTile() - 1), size, colour);
    }

    Colour get4TexelAverageColour(Vector topLeft)
    {
        return tileAt(topLeft>>texelsPerTileLevel())->
            get4TexelAverage(topLeft&(texelsPerTile() - 1));
    }

    void clearGrid()
    {
        removeActiveTiles();
        _gridSize = Vector(0, 0);
        _gridTiles.clear();
    }

    bool tilePresent(Vector tilePosition)
    {
        if (!tilePosition.inside(_gridSize))
            return false;
        return tileAt(tilePosition) != 0;
    }

    Direct3DTexturedTile*& tileAt(Vector tilePosition)
    {
        return _gridTiles[tilePosition.y*_gridSize.x+tilePosition.x];
    }

    Vector gridSize() { return _gridSize; }

    void removeTile(Vector tilePosition)
    {
        Direct3DTexturedTile* tile = tileAt(tilePosition);
        if (tile != 0) {
            _mipMapTileGrid->moveToReserve(tile);
            tileAt(tilePosition) = 0;
        }
    }

    Direct3DTexturedTile* createTile(Vector tilePosition)
    {
        Direct3DTexturedTile* tile = _mipMapTileGrid->createTile();
        tileAt(tilePosition) = tile;
        _activeTiles.add(tile);
        return tile;
    }

    void coarseHalve()
    {
        Vector newSize = _gridSize>>1;
        for (int y=0; y<newSize.y; ++y)
            for (int x=0; x<newSize.x; ++x) {
                Vector destPosition(x, y);
                Vector sourcePosition = destPosition<<1;
                Direct3DTexturedTile* dest = tileAt(destPosition);
                bool gotAny = false;
                for (int yy=0; yy<2; ++yy)
                    for (int xx=0; xx<2; ++xx)
                        if (tileAt(sourcePosition + Vector(xx, yy)) != 0)
                            gotAny = true;
                if (!gotAny) {
                    if (dest != 0)
                        removeTile(destPosition);
                    continue;
                }
                if (dest == 0)
                    dest = createTile(destPosition);
                for (int yy=0; yy<2; ++yy)
                    for (int xx=0; xx<2; ++xx) {
                        Vector quadrant(xx, yy);
                        Vector position = sourcePosition + quadrant;
                        Direct3DTexturedTile* source = tileAt(position);
                        if (source == 0) {
                            source = createTile(position);
                            _mipMapTileGrid->plotTile(position);
                        }
                        dest->coarseHalveFrom(source, quadrant);
                    }
            }
        moveGrid(newSize, Vector(0, 0));
    }
private:
    void removeActiveTiles()
    {
        Direct3DTexturedTile* nextTile;
        for (Direct3DTexturedTile* tile = _activeTiles.getNext(); tile != 0;
            tile = nextTile) {
            nextTile = _activeTiles.getNext(tile);
            _mipMapTileGrid->moveToReserve(tile);
        }
    }

    static int texelsPerTileLevel()
    {
        return Direct3DTexturedTile::texelsPerTileLevel();
    }

    static int texelsPerTile() { return 1<<texelsPerTileLevel(); }

    Vector _gridSize;

    std::vector<Direct3DTexturedTile*> _gridTiles;
    LinkedList<Direct3DTexturedTile> _activeTiles;

    Direct3DMipMapTileGrid* _mipMapTileGrid;
};


// This represents an infinitely zoomable and pannable image. It is composed of
// Direct3DTileGrid objects. _grids[0] is plotted to the screen and
// _grids[_n-1] is rendered to.
template<class FractalProcessor> class Direct3DMipMapTileGrid
{
    typedef Direct3DTileGrid<FractalProcessor> Direct3DTileGrid;
    typedef Screen<FractalProcessor> Screen;
public:
    Direct3DMipMapTileGrid(int n)
      : _n(n),
        _processor(0),
        _grids(n),
        _nTiles(0)
    {
        for (int i=n-1; i>=0; --i)
            _grids[i] = new Direct3DTileGrid(this);
    }

    ~Direct3DMipMapTileGrid()
    {
        for (int i=0; i<_n; ++i)
            delete _grids[i];
        deleteReserve();
    }

    void setProcessor(FractalProcessor* processor)
    {
        _processor = processor;
        _screen = processor->screen();
    }

    void setDevice(IDirect3DDevice9* device) { _device = device; }

    void paint() { _grids[0]->paint(); /*_processor->screen()->printZoomPoint();*/ }

    // Called after grow, shrink and move operations are complete to re-draw
    // tiles that need re-drawing and to update the pixel coordinates of tile
    // corners.
    void update()
    {
        Vector gridSize = _grids[0]->gridSize();
        for (int y=0; y<gridSize.y; ++y)
            for (int x=0; x<gridSize.x; ++x) {
                Vector tilePosition(x, y);
                if (!_screen->tileVisible(tilePosition)) {
                    // Remove tiles from all grids
                    int size = 1;
                    for (int i=0; i<_n; ++i) {
                        Vector p = tilePosition<<i;
                        for (int yy=0; yy<size; ++yy)
                            for (int xx=0; xx<size; ++xx)
                                _grids[i]->removeTile(p + Vector(xx, yy));
                        size <<= 1;
                    }
                    continue;
                }

                // See if we can double
                if (_grids[0]->tilePresent(tilePosition)) {
                    int size = 1;
                    for (int i=1; i<_n; ++i) {
                        size <<= 1;
                        Vector gridPosition = tilePosition << i;
                        if (_grids[i]->tilePresent(gridPosition))
                            continue;
                        for (int yy=0; yy<size; ++yy)
                            for (int xx=0; xx<size; ++xx) {
                                Vector p = gridPosition + Vector(xx, yy);
                                _grids[i]->createTile(p)->
                                    doubleFrom(_grids[i - 1]->
                                        tileAt(p >> 1), p & 1);
                            }
                    }
                    continue;
                }

                // See if we need to plot.
                int size = 1<<(_n - 1);
                for (int yy=0; yy<size; ++yy)
                    for (int xx=0; xx<size; ++xx) {
                        Vector p = (tilePosition<<(_n - 1)) + Vector(xx, yy);
                        if (_grids[_n - 1]->tilePresent(p))
                            continue;
                        _grids[_n - 1]->createTile(p);
                        _screen->plotTile(p);
                    }

                // See if we need to halve
                for (int i=_n-2; i>=0; --i) {
                    size >>= 1;
                    Vector gridPosition = tilePosition << i;
                    if (_grids[i]->tilePresent(gridPosition))
                        continue;
                    for (int yy=0; yy<size; ++yy)
                        for (int xx=0; xx<size; ++xx) {
                            Vector p = gridPosition + Vector(xx, yy);
                            Direct3DTexturedTile* tile = _grids[i]->
                                createTile(p);
                            Vector p2 = p<<1;
                            for (int j=0; j<4; ++j) {
                                Vector q(j&1, j>>1);
                                tile->halveFrom(_grids[i+1]->tileAt(p2|q), q);
                            }
                        }
                }
            }
        // Update tile vertices on level 0
        _grids[0]->updateVertices();
    }

    // After resizing is complete, delete tiles that we aren't using anymore
    // to reduce memory usage.
    void doneResize() { deleteReserve(); }

    // Set the pixels within the square whose top-left corner is "topLeft" and
    // whose edges are "size" texels to colour "colour". This square is assumed
    // to be entirely within a single tile. Called from plotTile().
    void plot(Vector topLeft, int size, Colour colour)
    {
        _grids[_n-1]->plotBlock(topLeft, size, colour);
    }

    // Called when a colour is decided to actually plot to the textures. Same 
    // as plot, but can be called without the assumption that the tile is 
    // present, and also updates the render tiles.
    void checkedPlot(Vector topLeft, int size, Colour colour)
    {
        // If the tile is absent, don't plot 
        if (!_grids[_n-1]->tilePresent(topLeft >> plotTexelsPerTileLevel()))
            return;

        // Update the mipmapped tiles as well
        for (int i=_n-1; i>=0; --i) {
            _grids[i]->plotBlock(topLeft, size, colour);
            if (i > 0) {
                if (size == 1)
                    colour = _grids[i]->get4TexelAverageColour(topLeft&(-2));
                else 
                    size >>= 1;
                topLeft >>= 1;
            }
        }
    }

    void moveToReserve(Direct3DTexturedTile* tile)
    {
        tile->remove();
        _reserveTiles.add(tile);
    }

    Direct3DTexturedTile* createTile()
    {
        Direct3DTexturedTile* tile = _reserveTiles.getNext();
        if (tile == 0) {
            tile = new Direct3DTexturedTile;
            tile->setDevice(_device);
            ++_nTiles;
        }
        else
            tile->remove();
        return tile;
    }

    // Halves the unit size of tiles when zooming in. Leaves the origin 
    // unchanged.
    void shrink()
    {
        Direct3DTileGrid* g = _grids[0];
        for (int i=0; i<_n-1; ++i)
            _grids[i] = _grids[i+1];
        _grids[_n-1] = g;
        g->clearGrid();
        g->moveGrid(_grids[_n-2]->gridSize() << 1, Vector(0, 0));
    }

    // Moves the origin along tile boundaries.
    void move(Vector topLeft, Vector bottomRight)
    {
        int mask = (1<<(_n-1))-1;
        Vector gridSize = ((bottomRight - topLeft) + mask)&~mask;
        for (int i=_n-1; i>=0; --i) {
            _grids[i]->moveGrid(gridSize, topLeft);
            gridSize = gridSize >> 1;
            topLeft >>= 1;
        }
    }

    // Bottom right tile
    Vector bottomRight()
    {
        return _grids[_n-1]->gridSize();
    }

    // Doubles the unit size of tiles when zooming out. Leaves the origin 
    // unchanged.
    void grow()
    {
        _grids[_n-1]->coarseHalve();

        Direct3DTileGrid* g = _grids[_n-2];
        for (int i=_n-2; i>=1; --i) {
            _grids[i] = _grids[i-1];
            _grids[i]->moveGrid(
                (_grids[i]->gridSize() + 1)&(-2), Vector(0, 0));
        }
        _grids[0] = g;
        g->clearGrid();
        g->moveGrid(_grids[1]->gridSize() >> 1, Vector(0, 0));
    }

    // Called by coarseHalve() to plot a missing source tile before halving.
    void plotTile(Vector tilePosition)
    {
        _screen->plotTile(tilePosition);
    }

    // Size of render-level tiles in plot-level texels
    int texelsPerTileLevel() const
    {
        return plotTexelsPerTileLevel() + _n - 1;
    }
    int texelsPerTile() const { return 1 << texelsPerTileLevel(); }

    // Called by Direct3DTileGrid to determine pixel coordinates. Note that the
    // texel coordinates here are grid 0 texel coordinates, hence the shift.
    Vector2<float> pixelFromTexel(Vector texel)
    {
        return _screen->pixelFromTexel(texel<<(_n - 1));
    }

    // Called by AnimatedWindow. We don't need to do anything here because with
    // Direct3DMipMapTileGrid, drawing happens in between window updates.
    void draw() { }

    // Called by ZoomingRotatingWindow to zoom, rotate and pan
    void zoomRotate(Fix16p16 zoomLevel, Fix16p16 angle, 
        Vector zoomPosition, bool panning)
    {
        _screen->zoomRotate(zoomLevel, angle, zoomPosition, panning, false);
    }

    // Called by ZoomingRotatingWindow to set the size of the window
    void resize(Vector newSize) { _screen->resize(newSize); }

private:
    void deleteReserve()
    {
        Direct3DTexturedTile* nextTile;
        for (Direct3DTexturedTile* tile = _reserveTiles.getNext(); tile != 0;
            tile = nextTile) {
            nextTile = _reserveTiles.getNext(tile);
            delete tile;
            --_nTiles;
        }
        _reserveTiles.clear();
    }

    // The size of the tiles in the plotting grid.
    static int plotTexelsPerTileLevel()
    {
        return Direct3DTexturedTile::texelsPerTileLevel();
    }
    static int plotTexelsPerTile() { return 1 << plotTexelsPerTileLevel(); }

    int _n;
    int _nTiles;
    FractalProcessor* _processor;
    Screen* _screen;
    std::vector<Direct3DTileGrid*> _grids;
    
    LinkedList<Direct3DTexturedTile> _reserveTiles;
    IDirect3DDevice9* _device;
};

#include "matrix.cpp"

// This handles the various coordinate systems:
// * Pixel coordinates (pixels)                       
// * Texel coordinates (texels)
// * Matrix coordinates (points)
// * Complex numbers (units)
// and transformations between them.
//
// Invariants:
// * All transformations are orthogonal, i.e. we assume square pixels and don't 
//   allow stretching by different amounts in different directions (though such
//   images could be generated by transformations of c before evaluation).
// * Texel coordinates, matrix coordinates and complex numbers cannot be 
//   rotated with respect to each other - the texel x axis and matrix x axis 
//   both correspond to the real axis and rotation is accomplished by rotating 
//   the pixel coordinates with respect to these.
// * Matrix coordinates are texel coordinates translated and scaled up by a 
//   power of 2 such that the entire matrix coordinate space is in the range
//   ([0,2^30), [0,2^30)).
// * Complex numbers are related to texel coordinates by a translation and a
//   scaling by a factor of 2 (so the distance from 0 to 1 corresponds to 2^m
//   point coordinates and 2^n texel coordinates for some integers m and n.
// * When we transform between any two of complex numbers, pixel coordinates
//   and matrix coordinates, we go via texel coordinates.
// * The screen covered by the tile grid, but would not be by a grid of
//   fewer tiles.
// * The top left tile has its top left corner at texel (0, 0).
// * The matrix covers all the tiles, but a matrix of 1/4 the size would
//   not.
// * log_2(_texelsPerPixel) is between _mipMapLevel-1 and _mipMapLevels.
// * _unitsPerTexel is an integer power of 2.
// * _originUnit is a multiple of _unitsPerTexel.
// * _matrixOrigin is on a tile boundary.
template<class FractalProcessor> class Screen
{
    typedef Direct3DMipMapTileGrid<FractalProcessor> Direct3DMipMapTileGrid;
    typedef Matrix<FractalProcessor> Matrix;
public:
    // Initialization
    Screen(int mipMapLevels)
      : _pixelsPerScreen(320, 240),
        _matrixLevel(30),
        _matrixOrigin(0, 0),
        _mipMapLevels(mipMapLevels),
        _angle(0),
        _zoomLevel(0),
        _originUnit(0, 0)
    { }

    //void printZoomPoint()
    //{
    //    printf("%.17lf + %.17lf i\n",_zoomUnit.x,_zoomUnit.y);
    //}

    void setProcessor(FractalProcessor* processor)
    { 
        _processor = processor;
        _tileGrid = processor->tileGrid();
        _matrix = processor->matrix();
        _texelsPerTileLevel = _tileGrid->texelsPerTileLevel();
        _matrixLevel -= _texelsPerTileLevel;
    }

    void setInitial(Complex<double> centre, double diagonal)
    {
        _zoomLevel0 = Fix16p16(-log(diagonal));
        _zoomPixel = Vector2<float>(0, 0);
        Vector2<double> pixelsPerScreen(_pixelsPerScreen);
        double pixelDiagonal = sqrt(pixelsPerScreen.modulus2());
        _zoomUnit = centre - Complex<double>(pixelsPerScreen)*diagonal/
            (2.0*pixelDiagonal);

        double unitsPerPixel = diagonal/pixelDiagonal;
        int level;
        frexp(unitsPerPixel, &level);
        _unitsPerTexel = ldexp(1.0, level-_mipMapLevels);
        _texelsPerPixel = static_cast<float>(unitsPerPixel/_unitsPerTexel);

        update(true);
    }

    // Called by the window to set the size of screen in pixel coordinates
    void resize(Vector pixelsPerScreen)
    {
        if (pixelsPerScreen.zeroArea() || pixelsPerScreen==_pixelsPerScreen)
            return;
        _zoomUnit = unitFromPixel(
            Vector2_static_cast<float>(_pixelsPerScreen)/2);
        _zoomPixel = Vector2_static_cast<float>(pixelsPerScreen)/2;
        _pixelsPerScreen = pixelsPerScreen;
        update(false);
    }

    // Zoom function for ZoomingRotatingWindow (position and magnification)
    // zoomLevel is a number indicating how far we are zoomed in (0 initially)
    // angle is in units of 2*pi radians (i.e. a number between 0 and 1, the
    // fraction of a full rotation that we have rotated anticlockwise).
    // zoomPixel is the pixel into which we are zooming and/or around which we
    // are rotating. If it stays the same, or if panning is true, the 
    // complex number corresponding to this pixel should not change.

    void zoomRotate(Fix16p16 zoomLevel, Fix16p16 angle, Vector zoomPixel,
        bool panning, bool initializing)
    {
        if (angle == _angle && zoomLevel == _zoomLevel &&
            zoomPixel == Vector2_static_cast<int>(_zoomPixel)) {
            // Nothing's changed - nothing to do.
            return;
        }

        if (!panning && zoomPixel != Vector2_static_cast<int>(_zoomPixel))
            _zoomUnit = unitFromPixel(Vector2_static_cast<float>(zoomPixel));
        _angle = angle;
        _zoomPixel = Vector2_static_cast<float>(zoomPixel);
        _zoomLevel = zoomLevel;
        update(false);
    }
        
    // Determines if any part of a particular block is on screen. The block in
    // question has its top-left corner at topLeftPoint and is 1<<sizeLevel 
    // points wide and high.
    bool blockVisible(Location location) const
    {
        return texelAreaVisible(texelFromPoint(location._topLeft),
            texelSizeFromPointLevel(location._size));
    }

    // Determines if a particular block is entirely on screen.
    bool blockEntirelyVisible(Location location) const
    {
        Vector texel = texelFromPoint(location._topLeft);
        int size = texelSizeFromPointLevel(location._size);
        if (!texelVisible(texel))
            return false;
        texel.x += size;
        if (!texelVisible(texel))
            return false;
        texel.y += size;
        if (!texelVisible(texel))
            return false;
        texel.x -= size;
        if (!texelVisible(texel))
            return false;
        return true;
    }

    // Returns the minimum block size (used to determine if a block is 
    // splittable)
    int smallPointSize() const { return _matrixLevel; }

    double delta2() const { return _unitsPerTexel*_unitsPerTexel; }

    // Converts matrix coordinates to complex numbers
    Complex<double> cFromPoint(Vector point) const
    {
        return cFromTexel(Vector2_static_cast<float>(texelFromPoint(point)));
    }

    // Whether or not a particular tile is visible (needs to be present in the
    // tile grid).
    bool tileVisible(Vector tilePosition) const
    {
        return texelAreaVisible(
            tilePosition << _texelsPerTileLevel, 1 << _texelsPerTileLevel);
    }

    // The pixel corresponding to a particular texel
    Vector2<float> pixelFromTexel(Vector texel) const
    {
        return (Vector2_static_cast<float>(texel) / _rotor) / _texelsPerPixel
            + _originPixel;
    }

    // Called by Direct3DMipMapTileGrid to re-plot a tile.
    void plotTile(Vector tilePosition)
    {
        _matrix->plot(Location(pointFromTexel(texelFromTile(tilePosition)),
            tilePointSize()));
    }

    // Called by Manipulator during post-iteration plotting to plot a block
    // which may not be in a tile.
    void checkedPlot(Location location, int colour)
    {
        _tileGrid->checkedPlot(texelFromPoint(location._topLeft),
            texelSizeFromPointLevel(location._size), MColour(colour));
    }

    // Called by Manipulator to plot a block without checking - used in tile
    // plotting.
    void plotBlock(Location location, int colour)
    {
        _tileGrid->plot(texelFromPoint(location._topLeft),
            texelSizeFromPointLevel(location._size), MColour(colour));
    }

    // Point size of a plotting tile. We don't allow any point blocks in the
    // matrix to be larger than this.
    int tilePointSize()
    {
        return _matrixLevel + _texelsPerTileLevel - (_mipMapLevels - 1);
    }

private:
    int texelSizeFromPointLevel(int pointLevel) const
    {
        return 1<<(pointLevel - _matrixLevel);
    }

    // The point corresponding to a particular texel
    Vector pointFromTexel(Vector texel) const
    {
        return (texel - _matrixOrigin)<<_matrixLevel;
    }

    // The texel corresponding to a particular matrix point (for plotting
    // purposes).
    Vector texelFromPoint(Vector point) const
    {
        return _matrixOrigin + (point>>_matrixLevel);
    }

    Vector texelFromTile(Vector tile) const
    {
        return tile << (_texelsPerTileLevel - (_mipMapLevels - 1));
    }

    Complex<double> cFromTexel(Vector2<float> texel) const
    {
        return _originUnit + Complex<double>(texel)*_unitsPerTexel;
    }

    // The texel corresponding to a particular pixel
    Vector2<float> texelFromPixel(Vector2<float> pixel) const
    {
        return (pixel - _originPixel)*_rotor*_texelsPerPixel;
    }
   
    // Returns the complex number corresponding to a particular pixel. This is
    // used for finding the point to zoom into based on the mouse position.
    Complex<double> unitFromPixel(Vector2<float> pixel) const
    {
        return cFromTexel(texelFromPixel(pixel));
    }

    // Whether or not a particular texel is on the screen
    bool texelVisible(Vector texel) const
    {
        return Vector2_static_cast<int>(pixelFromTexel(texel)).
            inside(_pixelsPerScreen);
    }

    // Determines if any part of a particular block is on screen. The block in
    // question has its top-left corner at texel and is size texels wide and
    // high. This needs to be as fast as possible as it will potentially be
    // called for every block in the matrix during a
    // splitLargePoints() operation.
    bool texelAreaVisible(Vector texel, int size) const
    {
        // Most blocks are visible, so do a quick crude visibility check first.
        // See if any part of the block is inside the inner bounding box.
        if (texel.x + size > _innerTopLeft.x &&
            texel.y + size > _innerTopLeft.y &&
            texel.x < _innerBottomRight.x &&
            texel.y < _innerBottomRight.y)
            return true;

        // See if any of the corners of the block are on the screen.
        if (texelVisible(texel))
            return true;
        texel.x += size;
        if (texelVisible(texel))
            return true;
        texel.y += size;
        if (texelVisible(texel))
            return true;
        texel.x -= size;
        if (texelVisible(texel))
            return true;
        texel.y -= size;

        // See if any of the corners of the screen are in the block.
        Vector pixel(0, 0);
        if (pixelInBlock(pixel, texel, size))
            return true;
        pixel.x += _pixelsPerScreen.x;
        if (pixelInBlock(pixel, texel, size))
            return true;
        pixel.y += _pixelsPerScreen.y;
        if (pixelInBlock(pixel, texel, size))
            return true;
        pixel.x -= _pixelsPerScreen.x;
        if (pixelInBlock(pixel, texel, size))
            return true;

        return false;
    }

    // Perform matrix expansion on zoom out
    void expandMatrix()
    {
        // Bottom-right texel of the tile grid.
        Vector tileBottomRight =_tileGrid->bottomRight()<<
            (_texelsPerTileLevel + 1 - _mipMapLevels);
        do {
            // Compute texel coordinates of current matrix bottom right
            int matrixTexels = 1<<(30 - _matrixLevel);
            Vector matrixBottomRight(_matrixOrigin + matrixTexels);

            // See if we need to expand in any direction
            bool expandLeft = _matrixOrigin.x > 0;
            bool expandUp = _matrixOrigin.y > 0;
            bool expandRight = tileBottomRight.x > matrixBottomRight.x;
            bool expandDown = tileBottomRight.y > matrixBottomRight.y;
            if (expandUp && !expandLeft)
                expandRight = true;
            if (expandDown && !expandRight)
                expandLeft = true;
            if (!expandLeft && !expandRight)
                break;
            Vector quadrant(0, 0);
            if (expandLeft)
                quadrant.x = 1;
            if (expandUp)
                quadrant.y = 1;
            _matrixOrigin -= quadrant<<(30 - _matrixLevel);
            --_matrixLevel;
            _processor->grow(quadrant);
            _matrix->grow(quadrant);
        } while (true);
    }

    void update(bool initializing)
    {
        // Acquire the lock so threads don't acquire c, plot or modify the grid
        // while the coordinate transformations are in flux.
        Lock lock(_processor);

        // Compute scale factors. The zoom level is defined such that the 
        // window diagonal is exp(-(_zoomLevel+_zoomLevel0)) units.
        double unitsPerNewScreen = exp(-(_zoomLevel+_zoomLevel0).toDouble());
        double newPixelsPerNewScreen =
            sqrt(static_cast<double>(_pixelsPerScreen.modulus2()));
        double unitsPerNewPixel = unitsPerNewScreen/newPixelsPerNewScreen;

        // Update pixel conversion factors to convert between new pixels and
        // old texels.
        _texelsPerPixel = static_cast<float>(unitsPerNewPixel/_unitsPerTexel);
        _rotor = Rotor2<float>(static_cast<float>(_angle.toDouble()));
        _originPixel = Vector2_static_cast<float>(_zoomPixel) +
            Vector2_static_cast<float>(Vector2<double>((_originUnit -
                _zoomUnit)/unitsPerNewPixel))/_rotor;

        // Find the corners of the new window in old texels.
        Vector a = Vector2_static_cast<int>(texelFromPixel(
            Vector2_static_cast<float>(0, 0)));
        Vector b = Vector2_static_cast<int>(texelFromPixel(
            Vector2_static_cast<float>(_pixelsPerScreen.x, 0)));
        Vector c = Vector2_static_cast<int>(texelFromPixel(
            Vector2_static_cast<float>(_pixelsPerScreen)));
        Vector d = Vector2_static_cast<int>(texelFromPixel(
            Vector2_static_cast<float>(0, _pixelsPerScreen.y)));
        // From these, compute the bounding boxes in old texels. The outer
        // boxes are snapped to tile boundaries.
        int x[4]; x[0]=a.x; x[1]=b.x; x[2]=c.x; x[3]=d.x;
        int y[4]; y[0]=a.y; y[1]=b.y; y[2]=c.y; y[3]=d.y;
        std::sort(&x[0], &x[4]);
        std::sort(&y[0], &y[4]);
        int texelsPerTile = 1 << _texelsPerTileLevel;
        int mask = (texelsPerTile<<(_mipMapLevels-1)) - 1;
        Vector tileTopLeft = Vector(x[0], y[0]) & ~mask;
        _innerTopLeft = Vector2<int>(x[1], y[1]);
        _innerBottomRight = Vector2<int>(x[2], y[2]);
        Vector tileBottomRight = (Vector(x[3], y[3]) + mask) & ~mask;

        // Do zoom-ins. These do not affect the texel origin but do change the
        // scale factors.
        while (_texelsPerPixel < (1<<(_mipMapLevels-1))) {
            --_matrixLevel;
            _matrix->splitMultiTileLeaves();
            _tileGrid->shrink();
            tileTopLeft <<= 1;
            _innerTopLeft <<= 1;
            _innerBottomRight <<= 1;
            tileBottomRight <<= 1;
            _matrixOrigin <<= 1;
            _unitsPerTexel /= 2.0;
            _texelsPerPixel *= 2.0;
        }

        // Do zoom-outs. These do affect the texel origin as well as the scale
        // factors, because we need to keep the matrix origin at a tile
        // boundary.
        while (_texelsPerPixel >= (1<<_mipMapLevels)) {
            _tileGrid->grow();
            ++_matrixLevel;
            _matrix->consolidateSmallLeaves();
            _matrixOrigin >>= 1;
            tileTopLeft >>= 1;
            _innerTopLeft >>= 1;
            _innerBottomRight >>= 1;
            tileBottomRight >>= 1;
            _unitsPerTexel *= 2.0;
            _texelsPerPixel /= 2.0;
            if (_texelsPerPixel >= (1<<_mipMapLevels))
                expandMatrix();
        }

        // Adjust the origins, keeping the matrix origin at a tile boundary.
        Vector offset = (tileTopLeft - _matrixOrigin)&mask;
        tileTopLeft -= offset;

        _tileGrid->move(tileTopLeft>>(_texelsPerTileLevel + 1 - _mipMapLevels),
            tileBottomRight>>(_texelsPerTileLevel + 1 - _mipMapLevels));
        _matrixOrigin -= tileTopLeft;
        _originUnit += Complex<double>(tileTopLeft) * _unitsPerTexel;
        _originPixel +=
            Vector2_static_cast<float>(tileTopLeft)/_texelsPerPixel/_rotor;
        tileBottomRight -= tileTopLeft;
        _innerTopLeft -= tileTopLeft;
        _innerBottomRight -= tileTopLeft;

        // Make the matrix cover an extra tile width down and right. Leaves in
        // this region won't be plotted but their existence prevents lines of 
        // very small tiles on the bottom and right edges of the matrix if
        // these edges are close to the edges of the screen.
        tileBottomRight += Vector(texelsPerTile, texelsPerTile);

        expandMatrix();

        // Perform matrix shrinking on zoom in
        do {
            Vector tl = pointFromTexel(Vector(0, 0)) >> 28;
            Vector br = pointFromTexel(tileBottomRight) >> 28;
            if (tl.x == br.x)
                if (br.x == 3)
                    --tl.x;
                else 
                    ++br.x;
            if (tl.y == br.y)
                if (br.y == 3)
                    --tl.y;
                else
                    ++br.y;
            if (br-tl == Vector(1, 1)) {
                _matrixOrigin += tl<<(28 - _matrixLevel);
                ++_matrixLevel;
                _matrix->shrink(tl);
                _processor->shrink(tl);
            }
            else
                break;
        } while (true);

        if (!initializing) {
            // Update the tile grid.
            _tileGrid->update();

            _matrix->consolidateOffScreenBlocks();

            // Even a tiny change may cause some more blocks to become visible
            // and therefore split, so we need to be sure that we check. Call
            // setProgressed() in order to attempt to split if there's nothing
            // else to do.
            _processor->setProgressed();
            _processor->resume();
        }
    }

    // Determines if a pixel is inside a square, given the texel coordinates of
    // the top left of the square and the size of the square in texels.
    bool pixelInBlock(Vector pixel, Vector topLeft, int size) const
    {
        return (Vector2_static_cast<int>(texelFromPixel(
            Vector2_static_cast<float>(pixel)))-topLeft).
                inside(Vector(size, size)); 
    }

    FractalProcessor* _processor;
    Direct3DMipMapTileGrid* _tileGrid;
    Matrix* _matrix;

    // The granularity of the texured tiles in texels
    int _texelsPerTileLevel;

    // Screen size in pixels
    Vector _pixelsPerScreen;

    // Rotation angle (from window).
    Fix16p16 _angle;

    // Zoom level (from window).
    Fix16p16 _zoomLevel;

    // Zoom pixel (from window).
    Vector2<float> _zoomPixel;

    // Complex number corresponding to _zoomPixel.
    Complex<double> _zoomUnit;

    // Conversion between texel coordinates and complex numbers
    Complex<double> _originUnit;  // complex number at the origin texel
    double _unitsPerTexel;  // Should always be a power of 2

    // Conversion between matrix coordinates and texel coordinates
    int _matrixLevel;  // log(points per texel)/log(2)
    Vector _matrixOrigin;  // texel coordinates of the matrix origin

    // Conversion between texel coordinates and pixel coordinates
    float _texelsPerPixel;
    Vector2<float> _originPixel;  // Pixel coordinates of the origin texel
    Rotor2<float> _rotor;

    // Texel coordinates of the inner bounding box of the screen.
    Vector _innerTopLeft;
    Vector _innerBottomRight;

    // Number of tile grid levels.
    int _mipMapLevels;

    // Initial zoom level.
    Fix16p16 _zoomLevel0;
};

// Window mixin that gives a window the capability to zoom in and out and pan
// using the mouse and rotate using the Z and X keys.
template<class Base> class ZoomingRotatingWindow : public Base
{
public:
    class Params
    {
        friend class ZoomingRotatingWindow;
    public:
        Params(typename Base::Params bp) : _bp(bp) { }
    private:
        typename Base::Params _bp;
    };

    ZoomingRotatingWindow(Params p)
      : Base(p._bp),
        _zoomPosition(0,0),
        _lButton(false),
        _rButton(false),
        _mButton(false),
        _zButton(false),
        _xButton(false),
        _zoomLevel(0),
        _zoomVelocity(0),
        _maxZoomVelocity(2048/65536.0),
        _zoomAcceleration(512/65536.0),
        _angle(0),
        _angularVelocity(0),
        _maxAngularVelocity(512/65536.0),
        _angularAcceleration(32/65536.0)
    { }

protected:
    virtual LRESULT handleMessage(UINT uMsg, WPARAM wParam, LPARAM lParam)
    {
        switch (uMsg) {
            case WM_LBUTTONDOWN:
                buttonDown(&_lButton, lParam);
                break;
            case WM_LBUTTONUP:
                buttonUp(&_lButton);
                break;
            case WM_RBUTTONDOWN:
                buttonDown(&_rButton, lParam);
                break;
            case WM_RBUTTONUP:
                buttonUp(&_rButton);
                break;
            case WM_MBUTTONDOWN:
                recalculateRegion(); // Get a fresh _zoomUnit.
                buttonDown(&_mButton, lParam);
                break;
            case WM_MBUTTONUP:
                buttonUp(&_mButton);
                break;
            case WM_MOUSEMOVE:
                zoomPos(VectorFromLParam(lParam));
                break;
            case WM_SIZE:
                onResize(VectorFromLParam(lParam));
                return 0;
            case WM_KILLFOCUS:
                _lButton = _rButton = _mButton = _zButton = _xButton = false;
                ReleaseCapture();
                break;
            case WM_KEYUP:
                if (wParam == 'Z')
                    _zButton = false;
                if (wParam == 'X')
                    _xButton = false;
                break;
            case WM_KEYDOWN:
                if (wParam == 'Z')
                    _zButton = true;
                if (wParam == 'X')
                    _xButton = true;
                if (wParam == VK_ESCAPE)
                    onDestroy();
                break;
        }

        return Base::handleMessage(uMsg, wParam, lParam);
    }

private:
    static Vector VectorFromLParam(LPARAM lParam)
    {
        return Vector(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
    }

    void buttonDown(bool* button, LPARAM lParam)
    {
        if (_lButton || _rButton || _mButton)
            zoomPos(VectorFromLParam(lParam));
        *button = true;
        if (_lButton || _rButton || _mButton)
            SetCapture(_hWnd);
    }
    void buttonUp(bool* button)
    {
        *button = false;
        if (!_lButton && !_rButton && !_mButton)
            ReleaseCapture();
    }

    void onResize(Vector s)
    {
        if (s.zeroArea())
            return;

        Base::onResize(s);
    }

    // Recalculate the region - called whenever we zoom or pan.
    void recalculateRegion()
    {
        _image->zoomRotate(_zoomLevel, _angle, _zoomPosition, _mButton);
    }

    // Recalculate the zoom level and (if necessary) the region
    void recalculateZoomRotate()
    {
        // Recalculate zoom
        if (!_lButton)
            if (!_rButton) {
                // No buttons pressed - slow to a stop
                if (_zoomVelocity > static_cast<Fix16p16>(0)) {
                    // Decelerate from zooming in
                    _zoomVelocity -= _zoomAcceleration;
                }
                if (_zoomVelocity < static_cast<Fix16p16>(0)) {
                    // Decelerate from zooming out
                    _zoomVelocity += _zoomAcceleration;
                }
            }
            else {
                // Right button pressed only - zoom out
                if (_zoomVelocity > -_maxZoomVelocity) {
                    // Accelerate to zoom out
                    _zoomVelocity -= _zoomAcceleration;
                }
            }
        else
            if (!_rButton) {
                // Left button pressed only - zoom in
                if (_zoomVelocity < _maxZoomVelocity) {
                    // Accelerate to zoom in
                    _zoomVelocity += _zoomAcceleration;
                }
            }
            else {
                // Both buttons pressed, keep same velocity
            }
        _zoomLevel += _zoomVelocity;

        // Recalculate rotation
        if (!_xButton)
            if (!_zButton) {
                // No buttons pressed - slow to a stop
                if (_angularVelocity > static_cast<Fix16p16>(0)) {
                    // Decelerate from rotating anticlockwise
                    _angularVelocity -= _angularAcceleration;
                }
                if (_angularVelocity < static_cast<Fix16p16>(0)) {
                    // Decelerate from rotating clockwise
                    _angularVelocity += _angularAcceleration;
                }
            }
            else {
                // X button pressed only - rotate clockwise
                if (_angularVelocity > -_maxAngularVelocity) {
                    // Accelerate clockwise
                    _angularVelocity -= _angularAcceleration;
                }
            }
        else
            if (!_zButton) {
                // X button pressed only - rotate anticlockwise
                if (_angularVelocity < _maxAngularVelocity) {
                    // Accelerate anticlockwise
                    _angularVelocity += _angularAcceleration;  
                }
            }
            else {
                // Both buttons pressed, keep same angular velocity
            }
        _angle += _angularVelocity;
        _angle &= _angle.frac();

        if (_angularVelocity != 0 || _zoomVelocity != 0 || _mButton)
            recalculateRegion();
    }

    // Change the zoom position. Called whenever the mouse is moved.
    void zoomPos(Vector s)
    {
        if (s == _zoomPosition)
            return;
        _zoomPosition = s;

        if (_mButton) {
            recalculateRegion();
            IF_ZERO_THROW(InvalidateRect(_hWnd, NULL, FALSE));
        }
    }

    virtual bool onPaint()
    {
        // Render the fractal data to the screen (only done when the timer has
        // expired, not on every redraw).
        if (_timerExpired) {
            // Update the coordinates for the next frame.
            recalculateZoomRotate();
        }

        // Paint the image to the window
        return Base::onPaint();
    }

    Vector _zoomPosition;
    bool _lButton, _rButton, _mButton, _zButton, _xButton;
    Fix16p16 _zoomLevel;
    Fix16p16 _zoomVelocity;
    Fix16p16 _maxZoomVelocity;
    Fix16p16 _zoomAcceleration;
    Fix16p16 _angle;
    Fix16p16 _angularVelocity;
    Fix16p16 _maxAngularVelocity;
    Fix16p16 _angularAcceleration;
};


template<class FractalProcessor> class Evaluator
{
    typedef IncompleteLeaf<FractalProcessor> IncompleteLeaf;
public:
    Evaluator(FractalProcessor* processor)
      : _processor(processor)
    {
        _bailoutRadius2 = processor->getBailoutRadius2();
        _maximumIterations = _processor->maximumIterations();
    }

    // Updates the IncompleteLeaf after evaluation
    void updateLeaf(IncompleteLeaf* leaf) const
    {
        leaf->_z = _z;
        leaf->_zS = _zS;
        leaf->_iterations += _maximumIterations;
    }

    void initFromLeaf(IncompleteLeaf* leaf, Complex<double> c,
        double delta2)
    {
        _z = leaf->_z;
        _zS = leaf->_zS;
        _c = c;
        _delta2 = delta2;
    }

    int result() const { return _result; }

protected:
    FractalProcessor* _processor;

    double _delta2;
    Complex<double> _c;
    Complex<double> _z;
    Complex<double> _zS;
    int _maximumIterations;
    double _bailoutRadius2;
    int _result;
};


template<class FractalProcessor> class MandelbrotEvaluator
  : public Evaluator<FractalProcessor>
{
public:
    MandelbrotEvaluator(FractalProcessor* processor)
      : Evaluator(processor)
    { }

    void iterate()
    {
        Complex<double> z = _z;
        Complex<double> zS = _zS;
        Complex<double> c = _c;
        int maximumIterations = _maximumIterations;
        double delta2 = _delta2;
        double bailoutRadius2 = _bailoutRadius2;

        for (int i=0; i<maximumIterations; i += 2) {
            double zr2 = z.x*z.x;
            double zi2 = z.y*z.y;
            z = Complex<double>(zr2 - zi2 + c.x, 2*z.x*z.y + c.y);
            if (zr2 + zi2 > bailoutRadius2) {
                _result = i+1;
                return;
            }

            zr2 = z.x*z.x;
            zi2 = z.y*z.y;
            z = Complex<double>(zr2 - zi2 + c.x, 2*z.x*z.y + c.y);
            if (zr2 + zi2 > bailoutRadius2) {
                _result = i+2;
                return;
            }

            zr2 = zS.x*zS.x;
            zi2 = zS.y*zS.y;
            zS = Complex<double>(zr2 - zi2 + c.x, 2*zS.x*zS.y + c.y);
            Complex<double> d = z - zS;
            if (d.modulus2() < delta2) {
                _result = 0;
                return;
            }
        }

        _z = z;
        _zS = zS;
        _result = -1;
    }

    bool inSet()
    {
        double r2 = _c.modulus2();
        double x = _c.x;
        if (r2*(8*r2-3) <= 3.0/32-x)
            return true;
        return r2+2*x+1 <= 1.0/16;
    }
};


template<class FractalProcessor, class Evaluator> class FractalThread
  : public Thread
{
    typedef IncompleteLeaf<FractalProcessor> IncompleteLeaf;
    typedef Manipulator<FractalProcessor> Manipulator;
    typedef WorkQueueList<FractalProcessor> WorkQueueList;
public:
    FractalThread(FractalProcessor* processor)
      : _ending(false),
        _failed(false),
        _processor(processor),
        _evaluator(processor),
        _running(false)
    {
        _screen = processor->screen();
        _matrix = processor->matrix();
        _queue = processor->queue();
        _ready.reset();
        _finished.reset();
    }

    // Signals the thread to come to an end.
    void end() { _ending=true; _ready.set(); }

    // Wait for a thread to finish if it hasn't already and rethrow the
    // exception on the main thread if it failed.
    void check()
    {
        _finished.wait();
        _finished.reset();
        if (_failed)
            throw _exception;
    }

    // Start the thread.
    void go()
    { 
        _finished.reset();
        _ready.set();
    }

    HANDLE finishedHandle() { return _finished; }

    void wait() { _finished.wait(); }

    void growMatrix(Vector quadrant)
    {
        Vector point(int(_point.x), int(_point.y));
        if ((point& 1) != Vector(0, 0))
            _pointValid = false;
        point = (point>>1) | (quadrant<<29);
        _point.x = point.x;
        _point.y = point.y;
    }

    void shrinkMatrix(Vector semiQuadrant)
    {
        Vector point(int(_point.x), int(_point.y));
        point = (point<<1) + (semiQuadrant<<28);
        if ((point & 0xc0000000) != Vector(0, 0))
            _pointValid = false;
        _point.x = point.x;
        _point.y = point.y;
    }

    bool progressed() const { return _progressed; }

	bool running() const { return _running; }

private:
    void postEvaluationProcessing()
    {
        // The block may have moved elsewhere while we were iterating due
        // to another thread completing an adjacent block. So, re-find the 
        // block by point coordinates again and make sure it's still
        // incomplete.
        if (!_pointValid) {
            // Matrix was grown or shrunk and this point didn't make the
            // cut (or we haven't done anything yet).
            return;
        }
        Vector point(int(_point.x), int(_point.y));
        Manipulator* manipulator = _matrix->manipulator(point);
        if (manipulator->isComplete()) {
            // This point got moved to a new IncompletePoint* and the other
            // thread got to it first.
            return;
        }
        if (manipulator->topLeft() != point) {
            // This point was consolidated away.
            return;
        }
        IncompleteLeaf* leaf = manipulator->incompleteLeaf();
        if (leaf->_iterations != _iterations) {
            // Some other thread got to it before us.
            return;
        }
        int i = _evaluator.result();
        if (i == 0) {
            // Point was found to be periodic.
            manipulator->complete(0);
            return;
        }
        if (i == -1) {
            // Leaf is still undecided after the maximum number of iterations
            // were completed.
            _evaluator.updateLeaf(leaf);
            if (leaf->_iterations > leaf->_colour)
                manipulator->plot(0);
            return;
        }
        // Point escaped.
        int s = manipulator->pointSize();

        manipulator->complete(i + _iterations);
    }

    // Returns true if we ran out of work to do.
    bool preEvaluationProcessing()
    {
        do {
            IncompleteLeaf* leaf = _processor->getNextLeaf();
            if (leaf == 0)
                return true;
            Manipulator* manipulator = _matrix->manipulator(leaf);
            _processor->setProgressed();
            Vector point = manipulator->topLeft();
            _point.x = point.x;
            _point.y = point.y;
            _evaluator.initFromLeaf(leaf, _screen->cFromPoint(point),
                _screen->delta2());
            _iterations = leaf->_iterations;
            if (_iterations == 0 && _evaluator.inSet()) {
                manipulator->complete(0);
                continue;
            }
            _pointValid = true;
            break;
        } while (true);
        return false;
    }

    void threadProc()
    {
        BEGIN_CHECKED {
            do {
                _pointValid = false;
                _ready.wait();
                _ready.reset();
                do {
                    {
                        Lock lock(_processor);
                        _running = true;
                        postEvaluationProcessing();
                        if (_ending) {
                            _running = false;
                            break;
                        }
                        if (preEvaluationProcessing()) {
                            _running = false;
                            break;
                        }
                    }
                    _evaluator.iterate();
                } while (true);
                _finished.set();
            } while (!_ending);
        } END_CHECKED(Exception& e) {
            _exception = e;
            _failed = true;
            _finished.set();
        }
    }

    Exception _exception;
    volatile bool _failed;
    volatile bool _ending;
	volatile bool _running;
    Event _ready;    // Set by UI thread to trigger this thread to start.
    Event _finished; // Set by this thread to tell the UI thread we're done.
    Evaluator _evaluator;

    volatile Vector _point;
    volatile bool _pointValid;
    FractalProcessor* _processor;
    Screen<FractalProcessor>* _screen;
    Matrix<FractalProcessor>* _matrix;
    WorkQueueList* _queue;

    unsigned int _iterations;
};


template<class FractalProcessor> class WorkQueue
{
    typedef IncompleteLeaf<FractalProcessor> IncompleteLeaf;
public:
    WorkQueue()
      : _next(0)
    { }

    void add(IncompleteLeaf* leaf)
    { 
        _list.add(leaf);
        if (_next == 0)
            _next = leaf;
    }

    void remove(IncompleteLeaf* leaf)
    {
        if (leaf == _next)
            increment();
        leaf->remove();
    }

    IncompleteLeaf* get()
    {
        IncompleteLeaf* leaf = _next;
        if (_next != 0)
            increment();
        return leaf;
    }

    void reseatLeaf(IncompleteLeaf* oldLeaf, IncompleteLeaf* newLeaf)
    {
        if (_next == oldLeaf)
            _next = newLeaf;
    }

    void reset() { _next = _list.getNext(); }
    bool empty() const { return _list.empty(); }

private:
    void increment() { _next = _list.getNext(_next); }

    LinkedList<IncompleteLeaf> _list;
    IncompleteLeaf* _next;
};


template<class FractalProcessor> class WorkQueueList
{
    typedef IncompleteLeaf<FractalProcessor> IncompleteLeaf;
    typedef WorkQueue<FractalProcessor> WorkQueue;
public:
    WorkQueueList()
      : _n(0), _current(0), _offset(0)
    { }

    void setProcessor(FractalProcessor* processor)
    {
        _processor = processor;
        reset();
    }

    void add(IncompleteLeaf* leaf, int pointSize)
    {
        if (_n == 0)
            _offset = pointSize;
        int queue = pointSize - _offset;
        while (queue < 0) {
            _queues.insert(_queues.begin(), 0);
            _queues[0] = new WorkQueue();
            ++queue;
            ++_n;
            --_offset;
            ++_current;
        }
        while (queue >= _n) {
            _queues.push_back(new WorkQueue());
            ++_n;
        }
        _queues[queue]->add(leaf);
        if (queue > _current)
            _current = queue;
    }

    void remove(IncompleteLeaf* leaf, int pointSize)
    {
        int queue = pointSize - _offset;
        _queues[queue]->remove(leaf);
        if (queue == 0)
            while (_n>0 && _queues[queue]->empty()) {
                delete _queues[0];
                _queues.erase(_queues.begin());
                ++_offset;
                --_n;
                --_current;
            }
        else
            if (queue == _n-1)
                while (_n>0 && _queues[_n-1]->empty()) {
                    delete _queues[_n-1];
                    _queues.pop_back();
                    --_n;
                    if (_current == _n)
                        --_current;
                }
    }

    void reseatLeaf(IncompleteLeaf* oldLeaf, IncompleteLeaf* newLeaf,
        int pointSize)
    {
        _queues[pointSize - _offset]->reseatLeaf(oldLeaf, newLeaf);
    }

    IncompleteLeaf* get()
    {
        while (!atEnd()) {
            IncompleteLeaf* leaf = _queues[_current]->get();
            if (leaf != 0)
                return leaf;
            _queues[_current]->reset();
            --_current;
        }
        return 0;
    }

    void reset() { _current = _n-1; }
    bool atEnd() const { return _current < 0; }
    void growMatrix() { --_offset; }
    void shrinkMatrix() { ++_offset; }

    ~WorkQueueList()
    {
        for (int i=0; i<_n; ++i) {
            WorkQueue* queue = _queues[i];
            if (queue != 0)
                delete queue;
        }
    }
private:
    int _n;
    FractalProcessor* _processor;
    std::vector<WorkQueue*> _queues;
    int _current;
    int _offset;
};


// This class allocates and frees memory.
//
// malloc() and free() have a severe problem for this application - they assume
// that a block, once allocated, cannot be moved. This causes a high degree of
// fragmentation in the common case that we allocate a large number of blocks
// and then remove most of them, since the remainder will cause most of the
// memory pages to remain committed.
//
// However, many of our blocks are quite easy to move, given just a pointer to
// it, because we know where all pointers that point to it are, and we know
// that no other threads will be accessing it. Also, the blocks of interest
// come in a small number of sizes (~20 in practice). This means it is
// practical to use a separate compacting heap for each block size, and keep
// these heaps compacted by moving the last block into the gap whenever a block
// is deallocated.
//
// These heaps grow in chunks of at least 1Mb and at least one block. Chunks
// need not be contiguous in memory, allowing the heaps to be interspersed.
template<class FractalProcessor> class Allocator
{
public:
    Allocator()
    {
        SYSTEM_INFO systemInfo;
        GetSystemInfo(&systemInfo);
        _granularity = systemInfo.dwAllocationGranularity;
    }

    void setProcessor(FractalProcessor* processor) { _processor = processor; }

    // A heap for a particular size of block. This is implemented using
    // VirtualAlloc() and VirtualFree() instead of HeapAlloc() and HeapFree()
    // because the latter set of functions never release address space.
    class Heap
    {
        // A Chunk is a unit of allocation of address space.
        class ChunkHeader
        {
        public:
            ChunkHeader* _previous;
        };

        // Each block holds one object. BlockHeaders always start on a word
        // boundary, so the least significant bits will always be 0. Hence we
        // can re-use the least significant bit for a deleted flag.
        class BlockHeader
        {
        public:
            BlockHeader* _previous;
        };

    public:
        // We need a default constructor to be able to put a Heap in a
        // std::map. However, default-constructed Heaps are never actually
        // used.
        Heap() { }

        Heap(FractalProcessor* processor, Allocator* allocator, int size)
          : _blockSize(size + _blockHeaderSize),
            _chunk(0),
            _freeBytes(0),
            _block(0),
            _processor(processor),
            _allocator(allocator)
        {
            // We avoid allocating chunks less than 1Mb, but we also round up
            // to a full number of chunks to avoid wasting space.
            _chunkSize = ((0x100000 - _chunkHeaderSize) / _blockSize + 1)
                * _blockSize + _chunkHeaderSize;

            // Round up to the granularity. With 64Kb granularity, we allocate
            // 1Mb+64Kb for all grid sizes up to 512*512 for 4-byte blocks and
            // 32*32 for 56-byte blocks. This means we rarely allocate chunks
            // of any other size, which helps keep the address space
            // defragmented.
            _chunkSize = _allocator->align(_chunkSize);

            // This is the number of spare bytes at the end of each chunk, due
            // to alignment requirements.
            _slackBytes = (_chunkSize - _chunkHeaderSize) % _blockSize;
        }

        // Allocates a new block and returns a pointer to it.
        template<class T> T* allocate()
        {
            BlockHeader* newBlock;
            if (_blockSize > _freeBytes) {
                // Allocate a new chunk. We'll let VirtualAlloc choose where to
                // put it.
                ChunkHeader* chunk = reinterpret_cast<ChunkHeader*>(
                    VirtualAlloc(NULL, _chunkSize, MEM_COMMIT | MEM_RESERVE,
                        PAGE_READWRITE));
                IF_ZERO_CHECK_THROW_LAST_ERROR(chunk);
                if (chunk == 0)
                    return 0;
                _processor->adjustMemory(_chunkSize);
                chunk->_previous = _chunk;
                _chunk = chunk;
                _freeBytes = _chunkSize - _chunkHeaderSize;
                newBlock = reinterpret_cast<BlockHeader*>(
                    reinterpret_cast<Byte*>(chunk) + _chunkHeaderSize);
            }
            else {
                // Allocate out of the current chunk.
                newBlock = reinterpret_cast<BlockHeader*>(
                    reinterpret_cast<Byte*>(_block) + _blockSize);
            }
            _freeBytes -= _blockSize;
            newBlock->_previous = _block;
            _block = newBlock;
            return object<T>();
        }
        
        // Deallocates the block at p. To avoid memory fragmentation and keep
        // memory usage to a minimum, the object highest in memory is moved to
        // p with a call to T::moveFrom().
        template<class T> void deallocate(T* p)
        {
            BlockHeader* block = reinterpret_cast<BlockHeader*>(
                reinterpret_cast<Byte*>(p) - _blockHeaderSize);
            if (block != _block) {
                T* old = object<T>();
                p->moveFrom(_processor, old);
            }
            _block = _block->_previous;
            _freeBytes += _blockSize;
            if (_freeBytes == _chunkSize - _chunkHeaderSize) {
                // We have an entirely empty chunk. Free it.
                ChunkHeader* oldChunk = _chunk;
                _chunk = _chunk->_previous;
                VirtualFree(oldChunk, 0, MEM_RELEASE);
                _freeBytes = _slackBytes;
                _processor->adjustMemory(-static_cast<int>(_chunkSize));
            }
        }

    private:
        template<class T> T* object() const
        {
            return reinterpret_cast<T*>(
                reinterpret_cast<Byte*>(_block) + _blockHeaderSize);
        }

        // Size of a block header.
        static const size_t _blockHeaderSize = 8;

        // Size of a chunk header.
        static const size_t _chunkHeaderSize = 8;

        // Size of a block.
        size_t _blockSize;
        
        // Size of a chunk.
        size_t _chunkSize;

        // Pointer to the chunk containing the most recently allocated block.
        ChunkHeader* _chunk;

        // Pointer to the most recently allocated block.
        BlockHeader* _block;

        // Number of bytes free in the current chunk.
        size_t _freeBytes;

        // Number of bytes free in a full chunk.
        size_t _slackBytes;

        FractalProcessor* _processor;

        Allocator* _allocator;
    };

    // Return a the heap that should be used for allocating blocks of size
    // "size".
    Heap* heapForSize(int size)
    {
        std::map<int, Heap>::iterator i = _heaps.find(size);
        if (i != _heaps.end())
            return &i->second;
        _heaps[size] = Heap(_processor, this, size);
        return &_heaps[size];
    }

    // Round a value up to the allocation granularity
    size_t align(size_t value)
    { 
        return (value + _granularity - 1) & ~(_granularity - 1);
    }

private:
    // Allocation granularity. We always allocate multiples of this size, and
    // allocations are aligned to this size.
    size_t _granularity;

    std::map<int, Heap> _heaps;
    FractalProcessor* _processor;
};


class FractalProcessor : public IdleProcessor, public Mutex
{
    typedef MandelbrotEvaluator<FractalProcessor> MandelbrotEvaluator;
    typedef FractalThread<FractalProcessor, MandelbrotEvaluator> FractalThread;
    typedef IncompleteLeaf<FractalProcessor> IncompleteLeaf;
    typedef Matrix<FractalProcessor> Matrix;
    typedef Screen<FractalProcessor> Screen;
    typedef Direct3DMipMapTileGrid<FractalProcessor> Direct3DMipMapTileGrid;
    typedef WorkQueueList<FractalProcessor> WorkQueueList;
    typedef Allocator<FractalProcessor> Allocator;
    typedef Grid<FractalProcessor> Grid;
public:
    FractalProcessor(Complex<double> centre, float zoomLevel)
      : _mipMapLevels(2),
        _screen(_mipMapLevels),
        _tileGrid(_mipMapLevels),
        _nThreads(0)
    {
        // Set a limit on how much memory we allocate (80% of physical RAM).
        // Because we're continually touching all our memory, allocating too
        // much causes everything else to get swapped out, and Windows becomes
        // unbearably slow.
        MEMORYSTATUS ms;
        GlobalMemoryStatus(&ms);
        _memoryLimit = (ms.dwTotalPhys / 5) * 4;
        _memoryUsed = 0;

        // Set up allocators.
        _allocator.setProcessor(this);
        for (BlockType t = gridBlockType; t != lastBlockType;
            t = static_cast<BlockType>(static_cast<int>(t) + 1))
            for (int i=0;i<16;++i)
                _heap[t][i] = _allocator.heapForSize(Grid(i, t).bytes());

        // Count available threads
        DWORD_PTR pam, sam;
        IF_ZERO_THROW(GetProcessAffinityMask(GetCurrentProcess(), &pam, &sam));
        for (DWORD_PTR p=1; p!=0; p<<=1)
            if ((pam&p) != 0)
                ++_nThreads;

        // Create threads
        _threads.resize(_nThreads);
        _threadHandles.resize(_nThreads);
        for (int i=0; i<_nThreads; ++i) {
            _threads[i] = new FractalThread(this);
            _threadHandles[i] = _threads[i]->finishedHandle();
            _threads[i]->start();
        }

        // Link up objects
        _screen.setProcessor(this);
        _tileGrid.setProcessor(this);
        _matrix.setProcessor(this);
        _queue.setProcessor(this);
        _screen.setInitial(centre, zoomLevel);
    }

    ~FractalProcessor()
    {
        // Tell the threads to stop what they're doing.
        for (int i=0; i<_nThreads; ++i)
            _threads[i]->end();
        // Wait for them all to actually stop and delete them. Don't rethrow
        // any exceptions here.
        for (int i=0; i<_nThreads; ++i) {
            _threads[i]->join();
            delete _threads[i];
        }
    }

    // Resume all the threads if there might be anything for them to do.
    void resume()
    {
        if (_queue.atEnd()) {
            _queue.reset();
            _progressed = false;
        }
        for (int i=0;i<_nThreads;++i)
            if (!_threads[i]->running())
                _threads[i]->go();
    }

    void setProgressed() { _progressed = true; }

    // IdleProcessor function - called in the message loop when there are no 
    // more messages to process.
    void idle()
    {
        bool gotMessage = false;

        do {
            // Wait until either one of the threads stops or we get a message.
            DWORD raisedEvent = MsgWaitForMultipleObjects(_nThreads, 
                &_threadHandles[0], FALSE, INFINITE, QS_ALLINPUT);
            if (raisedEvent < WAIT_OBJECT_0 + _nThreads) {
                int thread = raisedEvent - WAIT_OBJECT_0;
                _threads[thread]->check();
                // A thread stopped because it had nothing to
                // do.
            }
            else
                gotMessage = true;
        } while (!gotMessage);
    }

    // Called by Screen when zooming out to make the matrix span a larger area
    void grow(Vector quadrant)
    {
        _queue.growMatrix();
        for (int i=0; i<_nThreads; ++i)
            _threads[i]->growMatrix(quadrant);
    }

    // Called by Screen when zooming in to make the matrix span a smaller area
    void shrink(Vector semiQuadrant)
    {
        _queue.shrinkMatrix();
        for (int i=0; i<_nThreads; ++i)
            _threads[i]->shrinkMatrix(semiQuadrant);
    }

    // Called by the worker threads to get data to work on.
    IncompleteLeaf* getNextLeaf()
    {
		if (_queue.atEnd()) {
            if (_progressed || _matrix.splitLargeLeaves()) {
                // We still have incomplete points. Do some more processing.
                resume();
            }
		}
        return _queue.get();
    }

    int maximumIterations() const { return 0x4000; }
    double getBailoutRadius2() const { return 16; }

    Grid* allocateGrid(int size, BlockType blockType)
    {
        Allocator::Heap* heap = _heap[blockType][size];
        Grid* grid = heap->allocate<Grid>();
        if (grid == 0)
            throw std::bad_alloc();
        return grid;
    }

    void reseatLeaf(IncompleteLeaf* oldLeaf, IncompleteLeaf* newLeaf,
        int pointSize)
    {
        queue()->reseatLeaf(oldLeaf, newLeaf, pointSize);
    }

    void deallocateGrid(Grid* grid)
    {
        _heap[grid->_blockType][grid->_size]->deallocate(grid);
    }

    bool canAllocate() const { return _memoryUsed < _memoryLimit; }
    void adjustMemory(int bytes) { _memoryUsed += bytes; }

    Matrix* matrix() { return &_matrix; }
    Screen* screen() { return &_screen; }
    Direct3DMipMapTileGrid* tileGrid() { return &_tileGrid; }
    WorkQueueList* queue() { return &_queue; }

private:
    Allocator _allocator;
    Allocator::Heap* _heap[lastBlockType][16];
    size_t _memoryLimit;
    size_t _memoryUsed;
    int _mipMapLevels;
    WorkQueueList _queue;
    Matrix _matrix;
    Screen _screen;
    Direct3DMipMapTileGrid _tileGrid;
    int _precision;
    int _derivatives;
    std::vector<HANDLE> _threadHandles;
    int _nThreads;
    std::vector<FractalThread*> _threads;
    bool _progressed;
};


INT APIENTRY WinMain(HINSTANCE hInst, HINSTANCE, LPSTR, INT nCmdShow)
{
    int result = 0;
    BEGIN_CHECKED {
        ComInitializer ci;
        Windows windows(hInst);
        Direct3D direct3D;

        FractalProcessor processor(-0.5, 5.0);

        Window::Params wp(&windows,
            L"Mandelbrot Set with grid tree - click to zoom",
            Vector(320, 240));
        typedef RootWindow<Window> RootWindow;
        RootWindow::Params rwp(wp);
        typedef Direct3DMipMapTileGrid<FractalProcessor> 
            Direct3DMipMapTileGrid;
        typedef Direct3DWindow<RootWindow, Direct3DMipMapTileGrid> ImageWindow;
        ImageWindow::Params iwp(rwp, processor.tileGrid(), direct3D);
        typedef AnimatedWindow<ImageWindow> AnimatedWindow;
        AnimatedWindow::Params awp(iwp, 60);
        typedef ZoomingRotatingWindow<AnimatedWindow> ZoomingRotatingWindow;
        ZoomingRotatingWindow::Params zwp(awp);
        ZoomingRotatingWindow window(zwp);

        window.show(nCmdShow);
        result = pumpMessages(&processor);
        windows.check();
    } END_CHECKED(Exception& e) {
        e.display();
    }
    return result;
}


int __cdecl main()
{
    return WinMain(GetModuleHandle(NULL), NULL, "", SW_SHOWNORMAL);
}