// This file implements the Matrix. This stores information about complete and
// incomplete points using as little time and space overhead as possible.
//
// In order to store large amounts of information in as little space as
// possible, we avoid using C++ polymorphism which would store type information
// with every leaf - when we have a square grid 2^n by 2^n points of all the
// same type, we store them in a grid, which has bits describing the type at
// the start followed by the actual data.
//
// All variables called "size" measure the size of a square, grid or block in
// logarithmic units: size n means 2^n by 2^n points or blocks (according to 
// context).
//
// A block is either a grid block or a leaf block. A grid block holds a grid
// which contains other blocks, all of the same type. A leaf block is either
// complete or incomplete.

enum BlockType
{
    gridBlockType,
    incompleteLeafType,
    completeLeafType,
    lastBlockType  // Keep this last (not actually used in Grid).
};


// Forward declarations
template<class FractalProcessor> class Grid;
template<class FractalProcessor> class BlockPointer;

// A block is a member of a grid and is either a point or a pointer to a
// sub-grid.
class Block
{
    // Block contains no member data directly - we must cast the this pointer
    // to extract the data.
};


// A block representing a sub-grid
// sizeof(GridBlock) == 4
template<class FractalProcessor> class GridBlock : public Block
{ 
public:
    Grid<FractalProcessor>* _grid;
};


// A leaf represents a particular complex number - an input to the fractal 
// algorithm which produces a colour as an output.
class Leaf : public Block
{
public:
    // Colour to plot. This may be different from _iterations, as blocks that
    // have been subdivided should be plotted in the same colour as the block
    // that they were subdivided from until they've iterated by more than that.
    unsigned int _colour;
};


// A leaf that has escaped or is periodic.
class CompleteLeaf : public Leaf
{ };


// A leaf for which the decision about whether or not it is in the set has not
// been made.
template<class FractalProcessor> class IncompleteLeaf
  : public Leaf, public LinkedListMember<IncompleteLeaf<FractalProcessor> >
{
    typedef Grid<FractalProcessor> Grid;
    typedef BlockPointer<FractalProcessor> BlockPointer;
public:
    // BlockPointer for this leaf.
    BlockPointer blockPointer() const
    {
        return BlockPointer(_parent, _parent->positionForIndex(
            (reinterpret_cast<const Byte*>(this) - 
                (reinterpret_cast<const Byte*>(_parent) + sizeof(Grid))) /
                    sizeof(IncompleteLeaf)));
    }

    void init(IncompleteLeaf* from)
    {
        _colour = from->_colour;
        _iterations = from->_iterations;
        _z = from->_z;
        _zS = from->_zS;
    }

    void init(int colour)
    {
        _colour = colour;
        _iterations = 0;
        _z = 0;
        _zS = 0;
    }

    // Current z values. _z is iterated twice for each time _zS is iterated,
    // and the two values are compared for periodicity checking.
    Complex<double> _z;
    Complex<double> _zS;

    unsigned int _iterations;

    // Which grid is this leaf in?
    Grid* _parent;
};


// A grid is a square array of blocks whose width is a power of 2.
// sizeof(Grid) == 12
template<class FractalProcessor> class Grid
{
    typedef GridBlock<FractalProcessor> GridBlock;
    typedef IncompleteLeaf<FractalProcessor> IncompleteLeaf;
    typedef BlockPointer<FractalProcessor> BlockPointer;
public:
    // The address of the gridBlock containing this grid.
    BlockPointer blockPointer() const
    {
        if (_parent == 0)
            return BlockPointer();
        return BlockPointer(_parent, _parent->positionForIndex(_index));
    }

    Block* blockAtPosition(Vector position)
    { 
        return blockAtIndex(indexForPosition(position)); 
    }

    Vector positionForIndex(int index)
    {
        return Vector(index&((1<<_size) - 1), index >> _size);
    }

    int blockBytes()
    {
        if (_blockType == completeLeafType)
            return sizeof(CompleteLeaf);
        if (_blockType == incompleteLeafType)
            return sizeof(IncompleteLeaf);
        return sizeof(GridBlock);
    }

    int bytes() { return itemOffset(count()); }

    void reparent(BlockPointer address)
    {
        _parent = address._grid;
        if (_parent != 0) {
            address.subGrid() = this;
            _index = _parent->indexForPosition(address._position);
        }
        else
            _index = 29;
    }

    Grid(int size, BlockType blockType)
      : _size(size),
        _blockType(blockType)
    { }

    int indexForPosition(Vector position)
    { 
        return (position.y<<_size) | position.x;
    }

    void moveFrom(FractalProcessor* processor, Grid* oldGrid)
    {
        switch (oldGrid->_blockType) {
            case gridBlockType:
                {
                    memcpy(this, oldGrid, oldGrid->bytes());
                    GridBlock* gridBlock =
                        static_cast<GridBlock*>(blockAtIndex(0));
                    int n = count();
                    for (int i=0; i<n; ++i) {
                        Grid* grid = (gridBlock++)->_grid;
                        if (grid != 0)
                            grid->_parent = this;
                    }
                }
                break;
            case incompleteLeafType:
                {
                    *this = *oldGrid;

                    // We need to know the point size of these leaves.
                    Grid* g = this;
                    int pointSize = 0;
                    while (g->_parent != 0) {
                        pointSize -= g->_size;
                        g = g->_parent;
                    }
                    pointSize += g->_index;

                    // Move the leaves. We do this one by one instead of with
                    // a memcpy() in order to get the LinkedList pointers
                    // correct.
                    IncompleteLeaf* incompleteLeaf =
                        static_cast<IncompleteLeaf*>(blockAtIndex(0));
                    IncompleteLeaf* oldLeaf = static_cast<IncompleteLeaf*>(
                        oldGrid->blockAtIndex(0));
                    int n = count();
                    for (int i=0; i<n; ++i) {
                        *incompleteLeaf = *oldLeaf;
                        processor->reseatLeaf(oldLeaf, incompleteLeaf,
                            pointSize);
                        incompleteLeaf->_parent = this;
                        (incompleteLeaf++)->reseat();
                        ++oldLeaf;
                    }
                }
                break;
            case completeLeafType:
                memcpy(this, oldGrid, oldGrid->bytes());
                break;

        }
        processor->matrix()->reseatGrid(oldGrid, this);
        if (_parent != 0)
            reinterpret_cast<GridBlock*>(_parent->blockAtIndex(_index))->_grid 
                = this;
    }

private:
    int count() const { return 1 << (_size << 1); }

    int itemOffset(int index)
    { 
        if (_blockType == incompleteLeafType) {
            // Round the size of Grid up to the nearest 8 bytes to get the
            // alignment right for IncompleteLeaf, which contains doubles.
            return ((sizeof(Grid) + 7) & ~7) + index*blockBytes(); 
        }
        return sizeof(Grid) + index*blockBytes(); 
    }

    Block* blockAtIndex(int index)
    { 
        return reinterpret_cast<Block*>(
            reinterpret_cast<Byte*>(this) + itemOffset(index)); 
    }

public:
    // 0 for 1x1, 1 for 2x2, 2 for 4x4 etc... up to 15 for 32768x32768
    int _size : 5;

    // Type of items in grid. Technically only 2 are needed so we can reclaim a
    // bit here if necessary, but the VS debugger gets confused about enums
    // using the most significant bit in a bitfield.
    BlockType _blockType : 3;

    // 24 bits spare for precision and SOI information.

    // Which grid is this grid in? (0 if this is the matrix grid or is
    // disconnected).
    Grid* _parent;

    // Where in the parent grid is this grid?
    // If _parent==0 then this is instead used for the point size of the blocks
    // in this grid.
    int _index;
};


// This class encapsulates a point location and size. The lowest _size bits of
// the components of _topLeft should always be 0.
class Location
{
public:
    Location()
      : _topLeft(0, 0),
        _size(29)
    { }

    Location(Vector topLeft, int size)
      : _topLeft(topLeft),
        _size(size)
    { }

    // Divide this square up into a 2^size by 2^size grid, and return the
    // location of the square at position "position" within that grid.
    Location child(Vector position, int size)
    {
        int s = _size - size;
        return Location(_topLeft | (position << s), s);
    }

    // Is point contained within this location?
    bool contains(Vector point)
    {
        return (point-_topLeft).inside(Vector(1<<_size, 1<<_size));
    }

    Vector _topLeft;
    int _size;
};


// The following class hierarchy exists just to break up the matrix
// functionality into chunks, each more high-level than the last.

// This class encapsulates a block within a grid.
template<class FractalProcessor> class BlockPointer
{
protected:
    typedef Grid<FractalProcessor> Grid;
    typedef GridBlock<FractalProcessor> GridBlock;
    typedef IncompleteLeaf<FractalProcessor> IncompleteLeaf;

public:
    BlockPointer()
      : _grid(0),
        _position(0, 0)
    { }

    BlockPointer(Grid* grid, Vector position = Vector(0, 0))
      : _grid(grid),
        _position(position)
    { }

    IncompleteLeaf* incompleteLeaf() const
    {
        return static_cast<IncompleteLeaf*>(leaf());
    }

    bool isComplete() const { return blockType() == completeLeafType; }

    Grid*& subGrid() { return gridBlock()->_grid; }

    Grid* _grid;

    Vector _position;

protected:
    void ascend() { *this = _grid->blockPointer(); }

    // Assuming blockType()==gridBlockType, move to a block in the child grid.
    void descend(Vector position = Vector(0, 0))
    {
        _grid = subGrid();
        _position = position;
    }

    CompleteLeaf* completeLeaf() const
    {
        return static_cast<CompleteLeaf*>(leaf());
    }

    unsigned int& colour() { return leaf()->_colour; }
    BlockType blockType() const { return _grid->_blockType; }
    bool atTop() const { return _grid == 0; }
    Block* block() const { return _grid->blockAtPosition(_position); }
    Leaf* leaf() const { return static_cast<Leaf*>(block()); }
    int gridSize() const { return _grid->_size; }
    int gridWidth() const { return 1 << gridSize(); }
    bool atEnd() const { return _position.y == gridWidth(); }
    GridBlock* gridBlock() const { return static_cast<GridBlock*>(block()); }
    Grid* subGrid() const { return gridBlock()->_grid; }
};


// This class encapsulates a square group of blocks.
template<class FractalProcessor> class BlockGroup
  : public BlockPointer<FractalProcessor>
{
protected:
    BlockGroup(Grid* grid)
      : BlockPointer(grid),
        _size(0)
    { }

    bool atLastQuadrant()
    {
        return gridSize() == 0 || ((_position >> _size) & 1) == Vector(1, 1);
    }

    void ascend() { BlockPointer::ascend(); _size = 0; }

    void descend(Vector position = Vector(0, 0))
    {
        BlockPointer::descend(position);
        _size = 0;
    }

    int groupWidth() const { return 1 << _size; }

    // Move the blocks from source to this group.
    void moveBlocks(BlockGroup source)
    {
        int destinationStride = gridWidth();
        int sourceStride = source.gridWidth();
        int width = groupWidth();
        if (blockType() == incompleteLeafType) {
            IncompleteLeaf* sourceLine = source.incompleteLeaf();
            IncompleteLeaf* destinationLine = incompleteLeaf();
            for (int y=0; y<width; ++y) {
                IncompleteLeaf* source = sourceLine;
                IncompleteLeaf* destination = destinationLine;
                for (int x=0; x<width; ++x)
                    (destination++)->init(source++);
                sourceLine += sourceStride;
                destinationLine += destinationStride;
            }
            return;
        }
        int bytes = _grid->blockBytes() << _size;
        if (blockType() == gridBlockType) {
            GridBlock* sourceLine = source.gridBlock();
            GridBlock* destinationLine = gridBlock();
            for (int y=0; y<width; ++y) {
                memcpy(destinationLine, sourceLine, bytes);
                GridBlock* destination = destinationLine;
                for (int x=0; x<width; ++x) {
                    (destination++)->_grid->reparent(*this);
                    ++_position.x;
                }
                _position.x -= width;
                ++_position.y;
                sourceLine += sourceStride;
                destinationLine += destinationStride;
            }
            _position.y -= width;
            return;
        }
        CompleteLeaf* sourceLine = source.completeLeaf();
        CompleteLeaf* destinationLine = completeLeaf();
        for (int y=0; y<width; ++y) {
            memcpy(destinationLine, sourceLine, bytes);
            sourceLine += sourceStride;
            destinationLine += destinationStride;
        }
    }

    int _size;
};


// This class encapsulates a BlockGroup and a Location.
template<class FractalProcessor> class Site
  : public BlockGroup<FractalProcessor>
{
public:
    Vector& topLeft() { return _location._topLeft; }

protected:
    Site(Grid* grid = 0)
      : BlockGroup(grid)
    { }

    void moveNext()
    {
        int blocks = groupWidth();
        int points = pointWidth();
        _position.x += blocks;
        topLeft().x += points;
        if (_position.x == gridWidth()) {
            _position.x = 0;
            _position.y += blocks;
            topLeft() += Vector(-(points << (gridSize()-_size)), points);
        }
    }

    void moveNextQuadrant()
    {
        int blocks = groupWidth();
        int points = pointWidth();
        if (((_position.x >> _size) & 1) == 0) {
            // Move right
            _position.x += blocks;
            topLeft().x += points;
        }
        else {
            // Move left and down
            _position += Vector(-blocks, blocks);
            topLeft() += Vector(-points, points);
        }
    }

    void ascend()
    {
        topLeft() -= _position << (pointSize() - _size);
        pointSize() += gridSize() - _size;
        BlockGroup::ascend();
    }

    void descend(Vector position = Vector(0, 0))
    {
        pointSize() -= _size + subGrid()->_size;
        topLeft() += position << pointSize();
        BlockGroup::descend(position);
    }

    void ascendQuadrant()
    {
        if (_size >= gridSize() - 1)
            ascend();
        else {
            ++_size;
            ++pointSize();
            _position &= -groupWidth();
            topLeft() &= -pointWidth();
        }
    }

    void descendQuadrant()
    {
        if (_size == 0) {
            descend();
            _size = gridSize() - 1;
            if (_size < 0)
                _size = 0;
            pointSize() += _size;
        }
        else {
            --_size;
            --pointSize();
        }
    }

    // Descend to the leaf block containing point "point".
    void descendToLeaf(Vector point)
    {
        while (blockType() == gridBlockType)
            descend((point - topLeft()) >> (pointSize() - subGrid()->_size));
    }

    void moveTo(Vector point)
    {
        do {
            if (_grid->_parent == 0) {
                _position = point >> pointSize();
                topLeft() = point & -pointWidth();
                break;
            }
            else
                ascend();
        } while (!_location.contains(point));
        descendToLeaf(point);
    }

    bool moveLeft()
    {
        Vector point = topLeft() - Vector(1, 0);
        if (point.x >= 0) {
            moveTo(point);
            return true;
        }
        return false;
    }

    bool moveRight()
    {
        Vector point = topLeft() + Vector(pointWidth(), 0);
        if (point.x < 0x40000000) {
            moveTo(point);
            return true;
        }
        return false;
    }

    bool moveUp()
    {
        Vector point = topLeft() - Vector(0, 1);
        if (point.y >= 0) {
            moveTo(point);
            return true;
        }
        return false;
    }

    bool moveDown()
    {
        Vector point = topLeft() + Vector(0, pointWidth());
        if (point.y < 0x40000000) {
            moveTo(point);
            return true;
        }
        return false;
    }

    // Returns false if this square has a leaf block of a size less than half
    // the size of this block on any side.
    bool canConsolidateBlock()
    {
        Site l = *this;
        l.descendToLeaf(topLeft());
        if (l.moveLeft()) {
            if (l.pointSize() < pointSize() - 1)
                return false;
            if (l.pointSize() == pointSize() - 1) {
                l.moveDown();
                if (l.pointSize() < pointSize() - 1)
                    return false;
            }
        }
        Site t = *this;
        t.descendToLeaf(topLeft());
        if (t.moveUp()) {
            if (t.pointSize() < pointSize() - 1)
                return false;
            if (t.pointSize() == pointSize() - 1) {
                t.moveRight();
                if (t.pointSize() < pointSize() - 1)
                    return false;
            }
        }
        Vector rPoint = topLeft() + Vector(pointWidth(), pointWidth() >> 1);
        if (rPoint.x < 0x40000000) {
            Site r = *this;
            r.moveTo(rPoint);
            if (r.pointSize() < pointSize() - 1)
                return false;
            if (r.pointSize() == pointSize() - 1) {
                r.moveUp();
                if (r.pointSize() < pointSize() - 1)
                    return false;
            }
        }
        Vector dPoint = topLeft() + Vector(pointWidth() >> 1, pointWidth());
        if (dPoint.y < 0x40000000) {
            Site d = *this;
            d.moveTo(dPoint);
            if (d.pointSize() < pointSize() - 1)
                return false;
            if (d.pointSize() == pointSize() - 1) {
                d.moveLeft();
                if (d.pointSize() < pointSize() - 1)
                    return false;
            }
        }
        return true;
    }

    bool detail()
    {
        unsigned int c = colour();
        Site r = *this;
        if (!r.moveRight() || r.colour() != c)
            return true;
        Site d = *this;
        if (!d.moveDown() || d.colour() != c)
            return true;
        return false;
    }

    void setGroupSize(int size)
    {
        pointSize() += size - _size;
        _size = size;
    }

    void setPosition(Vector position)
    {
        topLeft() += (position - _position) << (pointSize() - _size);
        _position = position;
    }
public:
    int& pointSize() { return _location._size; }
    int pointSize() const { return _location._size; }
    int pointWidth() const { return 1 << pointSize(); }

    Location _location;
};


// This class encapsulates local transformations on the matrix and other
// operations requiring the FractalProcessor.
template<class FractalProcessor> class Manipulator
  : public Site<FractalProcessor>
{
public:
    void plot(unsigned int newColour)
    {
        if (newColour == colour())
            return;
        colour() = newColour;
        _processor->screen()->checkedPlot(_location, newColour);
    }

    void complete(unsigned int newColour)
    {
        plot(newColour);
        setBlockType(completeLeafType);

        colour() = newColour;
        tryConsolidateGrid();

        int texelSize = _processor->screen()->smallPointSize();
        if (visible() && detail() && pointSize() > texelSize) {
            splitLeaf();

            // See if the up and left leaves can be split.
            Vector point = topLeft();
            if (moveLeft()) {
                if (isComplete() && pointSize() > texelSize && visible() &&
                    detail())
                    splitLeaf();
                moveTo(point);
            }
            if (moveUp()) {
                if (isComplete() && pointSize() > texelSize && visible() &&
                    detail())
                    splitLeaf();
            }
        }
    }

    bool visible() const
    {
        return _processor->screen()->blockVisible(_location);
    }

    // Called by Screen to plot a tile, and recursively to plot part of a tile.
    // Always acts on the entire grid, but assumes _size==0.
    void plot(Location location)
    {
        int l = location._size - pointSize();

        // Move to the top-left of the blocks of interest.
        Vector position = (location._topLeft - topLeft()) >> pointSize();

        if (l <= 0) {
            setPosition(position);
            // We're plotting a single block or part of one.
            if (blockType() == gridBlockType)
                child().plot(location);
            else
                _processor->screen()->plotBlock(location, colour());
            return;
        }
        // We're plotting multiple subblocks.
        int width = 1 << l;
        for (int y=0; y<width; ++y)
            for (int x=0; x<width; ++x) {
                setPosition(position | Vector(x, y));
                if (blockType() == gridBlockType)
                    child().plot(_location);
                else
                    _processor->screen()->plotBlock(_location, colour());
            }
    }

    // Create a new grid in this gridblock.
    Grid* createGrid(int size, BlockType blockType)
    {
        Grid* grid = _processor->allocateGrid(size, blockType);
        *grid = Grid(size, blockType);
        grid->reparent(*this);
        if (blockType == incompleteLeafType) {
            descend();
            while (!atEnd()) {
                IncompleteLeaf* leaf = incompleteLeaf();
                leaf->_parent = grid;
                _processor->queue()->add(leaf, pointSize());
                moveNext();
            }
            ascend();
        }
        return grid;
    }

    Manipulator child(Vector position = Vector(0, 0)) const
    {
        Manipulator child = *this; 
        child.descend(position);
        return child;
    }

    // Attempt to consolidate grid with its siblings. Updates grid and position
    // so that they continue to refer to the same block.
    void tryConsolidateGrid()
    {
        Manipulator other = parent();
        if (other.atTop()) {
            // Can't consolidate the matrix itself.
            return;
        }
        if (other._grid->_parent == 0) {
            // Can't consolidate the subgrids of the matrix.
            return;
        }

        // In order to consolidate, this block's grid must be the same size and
        // type as its sibling grids.

        int size = gridSize();
        BlockType type = blockType();
        other.ascendQuadrant();
        other.descendQuadrant();
        for (int i=0; i<4; ++i) {
            Grid* quarter = other.subGrid();
            if (quarter->_size != size || quarter->_blockType != type)
                return;
            other.moveNextQuadrant();
        }
        Vector newPosition = _position | 
            ((parent()._position & 1) << size);
        int oldSize = _size;
        ascend();
        ascendQuadrant();
        descendQuadrant();

        // Consolidation will happen. Make the parent grid 2x2.
        while (gridSize() > 1)
            split();

        // Create new grid and update position
        other = *this;
        other._grid = parent().createGrid(size + 1, type);
        // Zero the old grid's parent pointer, so the new grid pointer won't
        // get overwritten if the old grid moves during the following deletes.
        _grid->_parent = 0;
        _grid->_index = pointSize();
        other._position = Vector(0, 0);
        other._size = size;
        other.topLeft() = topLeft();
        other.pointSize() = pointSize();

        // Move the data and delete the old grids
        for (int i=0; i<4; ++i) {
            descend();
            setGroupSize(size);
            other.moveBlocks(*this);
            ascend();
            other.moveNext();
            deleteChildGrid();
            moveNext();
        }

        // Swap this with the other and then delete the other to ensure our 
        // grid pointer is updated.
        Manipulator other2 = *this;
        *this = other;
        other2.deleteGrid();

        setGroupSize(oldSize);
        setPosition(newPosition);

        tryConsolidateGrid();
    }

    // Deletes the grid in this gridBlock, and all sub-grids of that grid.
    void recursiveDeleteChildGrid()
    {
        int i = 0;
        do {
            if (subGrid()->_blockType == gridBlockType)
                descend();
            else
                do {
                    deleteChildGrid();
                    if (i == 0)
                        return;
                    --i;
                    moveNext();
                    if (!atEnd())
                        break;
                    ascend();
                } while (true);
            ++i;
        } while (true);
    }

    void deleteChildGrid()
    {
        // We need to clear the pointer in the grid before we delete, or we
        // might move this block while it's in a non-valid state. But we need
        // to save the _grid pointer first so we know what to delete. Hence
        // this slightly roundabout way of doing things.
        Manipulator c = child();
        subGrid() = 0;
        c.deleteGrid();
    }

    // Deletes the entire grid that this is the top-left block of.
    void deleteGrid()
    {
        if (blockType() == incompleteLeafType)
            while (!atEnd()) {
                _processor->queue()->remove(incompleteLeaf(), pointSize());
                moveNext();
            }
        _processor->deallocateGrid(_grid);
    }

protected:
    Manipulator()
      : _processor(0)
    { }

    void moveToLeaf(IncompleteLeaf* leaf)
    {
        BlockPointer pointer = leaf->blockPointer();
        *this =
            Manipulator(_processor, pointer._grid).child(pointer._position);
    }

    Manipulator parent() const
    {
        Manipulator parent = *this;
        parent.ascend();
        return parent;
    }

    // Splits this leaf block into 4. The top left new leaf will be the same
    // as the original leaf and the other 3 will be incomplete leaves. Leaves
    // the address pointing to the same (now smaller) leaf. Splits adjacent
    // leaves as necessary to maintain the adjacency invariant.
    void splitLeaf()
    {
        Manipulator l = *this;
        Vector point = topLeft();
        if (l.moveLeft() && l.pointSize() > pointSize()) {
            *this = l;
            splitLeaf();
            moveTo(point);
        }
        Manipulator r = *this;
        if (r.moveRight() && r.pointSize() > pointSize()) {
            *this = r;
            splitLeaf();
            moveTo(point);
        }
        Manipulator u = *this;
        if (u.moveUp() && u.pointSize() > pointSize()) {
            *this = u;
            splitLeaf();
            moveTo(point);
        }
        Manipulator d = *this;
        if (d.moveDown() && d.pointSize() > pointSize()) {
            *this = d;
            splitLeaf();
            moveTo(point);
        }
        uncheckedSplitLeaf();
    }

    void uncheckedSplitLeaf()
    {
        if (blockType() == incompleteLeafType) {
            IncompleteLeaf leaf;
            leaf.init(incompleteLeaf());
            setBlockType(gridBlockType, incompleteLeafType);
            IncompleteLeaf* newLeaf = incompleteLeaf();
            (newLeaf++)->init(&leaf);
            for (int i=1; i<4; ++i)
                (newLeaf++)->init(leaf._colour);
        }
        else {
            // Make a copy of the point
            CompleteLeaf leaf = *completeLeaf();
            setBlockType(gridBlockType, gridBlockType);
            createGrid(0, completeLeafType);
            descend();
            *completeLeaf() = leaf;
            ascend();
            moveNext();
            while (!atEnd())
                createLeaf(leaf._colour);
            setPosition(Vector(0, 0));
        }
        tryConsolidateGrid();
    }

    // Consolidates this leaf block with its siblings.
    void consolidateLeaf()
    {
        // A 1x1 grid is handled by consolidating its parent, which will be 
        // at least 2x2.
        if (gridSize() == 0)
            ascend();

        // In order to consolidate, need to reduce the grid to 2x2.
        while (gridSize() > 1)
            split();

        BlockType type = blockType();
        setPosition(Vector(0, 0));

        Grid* newGrid;
        if (type == gridBlockType) {
            // We are a grid of grids - we can just reparent.
            newGrid = child()._grid;
            //if (newGrid->_size != 0)
            //    printf("Consolidating non-leaf!\n");
            newGrid->reparent(parent());
            if (newGrid->_blockType == incompleteLeafType) {
                // We're changing the size of this leaf, so we need to move it
                // into a different queue.
                IncompleteLeaf* leaf = static_cast<IncompleteLeaf*>(
                    newGrid->blockAtPosition(Vector(0, 0)));
                WorkQueueList<FractalProcessor>* queue = _processor->queue();
                queue->remove(leaf, pointSize());
                queue->add(leaf, pointSize() + 1);
            }
            for (int i=1; i<4; ++i) {
                moveNext();
                recursiveDeleteChildGrid();
            }
        }
        else {
            newGrid = parent().createGrid(0, type);
            parent().child().moveBlocks(*this);
        }
        // This should not invalidate newGrid, because newGrid should be 1x1
        // and we're deleting a 2x2 grid.
        deleteGrid();
        _grid = newGrid;
        setPosition(Vector(0, 0));
        ++pointSize();

        tryConsolidateGrid();
    }

    // Removes the entire block and replaces it with its top-left leaf.
    void consolidateBlock()
    {
        if (_size > 0) {
            while (_size < gridSize())
                split();
            ascend();
        }

        // Make a copy of the leaf.
        Manipulator corner = *this;
        corner.descendToLeaf(topLeft());
        IncompleteLeaf incomplete;
        CompleteLeaf complete;
        BlockType type = corner.blockType();
        if (type == incompleteLeafType)
            incomplete.init(corner.incompleteLeaf());
        else
            complete = *corner.completeLeaf();

        // Replace the grid.
        recursiveDeleteChildGrid();
        createGrid(0, type);
        descend();

        // Copy the leaf back.
        if (type == incompleteLeafType)
            incompleteLeaf()->init(&incomplete);
        else
            *completeLeaf() = complete;

        tryConsolidateGrid();
    }

    // Creates and initializes a 1x1 incompleteLeaf grid. Used by grow(),
    // initMatrix() and splitLeaf().
    void createLeaf(int colour = 0)
    {
        createGrid(0, incompleteLeafType);
        child().incompleteLeaf()->init(colour);
        moveNext();
    }

    FractalProcessor* _processor;

    // Split grid into 4 and updates grid and position so that they continue to
    // refer to the same block.
    void split()
    {
        Vector originalTopLeft = topLeft();
        Vector oldPosition = _position;
        Location oldLocation = _location;
        int oldSize = _size;

        // Create new grid and move data into it.
        int size = gridSize() - 1;
        Grid* quad = parent().createGrid(1, gridBlockType);

        Manipulator dest = parent().child();
        BlockType type = blockType();
        setPosition(Vector(0, 0));
        setGroupSize(size);
        for (int i=0; i<4; ++i) {
            dest.createGrid(size, type);
            dest.descend();
            dest.setGroupSize(size);
            dest.moveBlocks(*this);
            dest.ascend();
            dest.moveNext();
            moveNext();
        }

        // Save this grid for later deletion.
        Manipulator other = *this;

        // Adjust grid and position.
        _grid = BlockPointer(quad, oldPosition >> size).subGrid();
        oldPosition &= ((1<<size) - 1);
        _position = oldPosition;
        _size = oldSize;
        _location = oldLocation;

        // Delete the previous grid
        other.setPosition(Vector(0, 0));
        other.setGroupSize(0);
        other.deleteGrid();

        ascend();
        tryConsolidateGrid();
        descend();

        setPosition(oldPosition);
        setGroupSize(oldSize);
    }

private:
    // Changes the type of the block to "type". If "type" is gridBlockType,
    // make it a 2x2 grid of "gridType".
    void setBlockType(BlockType type, BlockType gridType = gridBlockType)
    {
        if (blockType() == type) {
            // No change is necessary - nothing to do
            return;
        }
        // In order to change the type of a block, it needs to be in a 1x1
        // grid.
        while (gridSize() > 0)
            split();

        // Now we have a 1x1 grid, the change is easy - just create a new grid
        // for the new block.
        ascend();
        deleteChildGrid();
        if (type == gridBlockType)
            createGrid(1, gridType);
        else
            createGrid(0, type);
        descend();

        // Don't consolidate the grid here, as the new grid won't be
        // initialized. Calling functions will do it later.
    }

    // Manipulator for the GridBlock that holds Grid g.
    Manipulator(FractalProcessor* processor, Grid* g)
      : _processor(processor)
    {
        if (g->_parent == 0)
            _location = Location(Vector(0, 0), 30);
        else {
            BlockPointer blockPointer = g->blockPointer();
            _grid = blockPointer._grid;
            _position = blockPointer._position;
            _location = Manipulator(processor, _grid)._location.child(
                _position, gridSize());
        }
    }
};


// This class holds the root grid pointer and encapsulates global matrix
// operations.
template<class FractalProcessor> class Matrix : Manipulator<FractalProcessor>
{
public:
    Matrix()
      : _root(0)
    { }

    ~Matrix()
    { 
        reset();
        if (_root != 0) {
            while (!atEnd()) {
                recursiveDeleteChildGrid();
                moveNext();
            }
            deleteGrid();
        }
    }

    // Manipulator for a particular IncompleteLeaf.
    Manipulator* manipulator(IncompleteLeaf* leaf)
    {
        reset();
        moveToLeaf(leaf);
        return this;
    }

    // Manipulator for a particular point.
    Manipulator* manipulator(Vector topLeft)
    {
        reset();
        moveTo(topLeft);
        return this;
    }

    // Called by Screen to plot a tile.
    void plot(Location location)
    {
        reset();
        Manipulator::plot(location);
    }

    bool splitLargeLeaves()
    {
        reset();
        int size = _processor->screen()->smallPointSize();
        bool split = false;
        do {
            if (!visible()) {
                if (!atLastQuadrant()) {
                    moveNextQuadrant();
                    continue;
                }
                // Ascend
            }
            else
                if (blockType() == gridBlockType && _size == 0) {
                    descendQuadrant();
                    if (blockType() != incompleteLeafType &&
                        pointSize() - _size > size)
                        continue;
                    // Ascend
                }
                else
                    if (_size > 0) {
                        descendQuadrant();
                        continue;
                    }
                    else {
                        if (detail()) {
                            if (!_processor->canAllocate()) {
                                // We're low on memory - do some more iterating
                                // before splitting any more.
                                return split;
                            }
                            if (pointSize() - 1 > size)
                                size = pointSize() - 1;
                            split = true;
                            splitLeaf();
                            continue;
                        }
                        else {
                            if (!atLastQuadrant()) {
                                moveNextQuadrant();
                                continue;
                            }
                            // Ascend
                        }
                    }
            do {
                ascendQuadrant();
                if (atTop())
                    return split;
                if (atLastQuadrant())
                    continue;
                moveNextQuadrant();
                break;
            } while (true);
        } while (true);
    }

    // Splits leaves larger than a tile.
    void splitMultiTileLeaves()
    {
        reset();
        int size = _processor->screen()->tilePointSize();
        do {
            if (blockType() == gridBlockType && _size == 0) {
                descendQuadrant();
                if (pointSize() - _size > size)
                    continue;
                // Ascend
            }
            else
                if (_size > 0) {
                    descendQuadrant();
                    continue;
                }
                else {
                    uncheckedSplitLeaf();  // Don't need to split adjacents.
                    if (pointSize() > size)
                        continue;
                    ascendQuadrant();
                    if (!atLastQuadrant()) {
                        moveNextQuadrant();
                        continue;
                    }
                    // Ascend
                }
            do {
                ascendQuadrant();
                if (atTop())
                    return;
                if (atLastQuadrant())
                    continue;
                moveNextQuadrant();
                break;
            } while (true);
        } while (true);
    }

    // Consolidate leaves smaller than a texel.
    void consolidateSmallLeaves()
    {
        reset();
        int smallPointSize = _processor->screen()->smallPointSize();
        descendQuadrant();
        do {
            while (blockType() == gridBlockType || _size > 0)
                descendQuadrant();
            if (pointSize() < smallPointSize) {
                consolidateLeaf();
                continue;
            }
            do {
                ascendQuadrant();
                if (atTop())
                    return;
                if (atLastQuadrant())
                    continue;
                moveNextQuadrant();
                break;
            } while (true);
        } while (true);
    }

    void consolidateOffScreenBlocks()
    {
        reset();
        int tileSize = _processor->screen()->tilePointSize();
        do {
            if (!visible() && pointSize() <= tileSize && 
                (_size > 0 || blockType()==gridBlockType) &&
                canConsolidateBlock()) {
                consolidateBlock();
                if (!atLastQuadrant()) {
                    moveNextQuadrant();
                    continue;
                }
                // Ascend
            }
            else
                if (!_processor->screen()->blockEntirelyVisible(_location) &&
                    (blockType() == gridBlockType || _size > 0)) {
                    descendQuadrant();
                    continue;
                }
                else {
                    if (!atLastQuadrant()) {
                        moveNextQuadrant();
                        continue;
                    }
                    // Ascend;
                }
            do {
                ascendQuadrant();
                if (atTop())
                    return;
                if (atLastQuadrant())
                    continue;
                moveNextQuadrant();
                break;
            } while (true);
        } while (true);
    }


    // Creates the matrix and returns its 2x2 grid.
    void setProcessor(FractalProcessor* processor)
    {
        _processor = processor;
        _root = createGrid(1, gridBlockType);
        reset();
        while (!atEnd())
            createLeaf();
        splitMultiTileLeaves();
    }

    // Enlarge the matrix when zooming out. The original matrix will be in the
    // "quadrant" quadrant of the new matrix.
    void grow(Vector quadrant)
    {
        reset();
        _root = parent().createGrid(1, gridBlockType);
        _grid->reparent(BlockPointer(_root, quadrant));
        reset();

        // Create some new 1x1 IncompleteLeaf blocks for the other 3
        // quadrants.
        for (int i=0; i<4; ++i)
            if (_position != quadrant)
                createLeaf();
            else
                moveNext();
        setPosition(quadrant);

        child().child().tryConsolidateGrid();

        splitMultiTileLeaves();
    }

    // Shrink the matrix when zooming in. semiQuadrant is the top left of the
    // new matrix in units of 1/4 of the original matrix width. The new matrix
    // will have 1/4 of the area of the original matrix.
    void shrink(Vector semiQuadrant)
    {
        reset();
        // First make sure each subgrid of the grid is a 2x2 gridBlock grid.
        for (int i=0; i<4; ++i) {
            descend();
            // If it's a 1x1 leaf grid, split it up.
            if (gridSize() == 0)
                splitLeaf();
            // If it's not a 2x2 grid of grids, split it up.
            if (blockType() != gridBlockType || gridSize() != 1) {
                split();
                ascend();
            }
            ascend();
            moveNext();
        }

        // Pick the 1/16th grids we need.
        Grid** newQuadrants[4];
        setPosition(Vector(0, 0));
        for (int i=0; i<4; ++i) {
            descend();
            for (int j=0; j<4; ++j) {
                Vector p((j&1) | ((i&1)<<1), (j>>1) | (i&2));
                p -= semiQuadrant;
                if (p.inside(Vector(2, 2)))
                    newQuadrants[p.x | (p.y<<1)] = &subGrid();
                moveNext();
            }
            ascend();
            moveNext();
        }
        // Swap these grids with the ones in the top-left quadrant
        setPosition(Vector(0, 0));
        descend();
        for (int i=0; i<4; ++i) {
            // Swap the grids
            Grid* g1 = subGrid();
            Grid* g2 = *newQuadrants[i];
            subGrid() = g2;
            *newQuadrants[i] = g1;

            // Swap their parents
            Grid* parent = g1->_parent;
            g1->_parent = g2->_parent;
            g2->_parent = parent;

            // Swap their indexes
            int index = g1->_index;
            g1->_index = g2->_index;
            g2->_index = index;

            moveNext();
        }
        ascend();

        // Delete the other 3 quadrants
        for (int i=1; i<4; ++i) {
            moveNext();
            recursiveDeleteChildGrid();
        }

        // Make the top-left quadrant the root
        setPosition(Vector(0, 0));
        descend();
        _root = _grid;

        // Delete the previous root
        parent().deleteGrid();

        // Fix up the new root's parent
        _root->_parent = 0;
        _root->_index = 29;
    }

    void reseatGrid(Grid* oldGrid, Grid* newGrid)
    {
        if (_root == oldGrid)
            _root = newGrid;
        if (_grid == oldGrid)
            _grid = newGrid;
    }

private:

    // Resets the manipulator to the top-left quadrant of the matrix.
    void reset()
    {
        _grid = _root;
        _position = Vector(0, 0);
        _size = 0;
        topLeft() = Vector(0, 0);
        pointSize() = 29;
    }

    Grid* _root;
};