#ifndef AJLIB_USER_H
#define AJLIB_USER_H

#include "def.h"
#include <windowsx.h>
#include <vector>
#include "vectors.h"

class ScrollBar
{
public:
    ScrollBar(double min, double max, double step, double value)
      : _min(min),
        _max(max),
        _step(step),
        _value(value) { }
    void setControl(HWND hDlg, int nIDDlgItem)
    {
        _hCtl = GetDlgItem(hDlg, nIDDlgItem);
        IF_NULL_THROW(_hCtl);
        update();
    }
    double getValue() const { return _value; }
    bool onScroll(int cmd, int pos)
    {
        bool handled = true;
        switch (cmd) {
            case SB_LEFT:
                _value = _min;
                break;
            case SB_RIGHT:
                _value = _max;
                break;
            case SB_LINELEFT:
            case SB_PAGELEFT:
                _value -= _step;
                break;
            case SB_LINERIGHT:
            case SB_PAGERIGHT:
                _value += _step;
                break;
            case SB_THUMBTRACK:
                _value = _min+static_cast<double>(pos)*(_max-_min)/0xffff;
                break;
            case SB_ENDSCROLL:
            case SB_THUMBPOSITION:
                handled = false;
                break;
            default:
                handled = false;
                assertTrue(false, L"Unknown scrollbar message");
        }
        if (_value < _min)
            _value = _min;
        if (_value > _max)
            _value = _max;
        update();
        return handled;
    }
    HWND getHandle() const { return _hCtl; }

private:
    int scale(double v) { return static_cast<int>(0xffff*(v-_min)/(_max-_min)); }

    void update()
    {
        SCROLLINFO si;
        ZeroMemory(&si, sizeof(SCROLLINFO));
        si.cbSize = sizeof(SCROLLINFO);
        si.fMask = SIF_PAGE | SIF_POS | SIF_RANGE;
        si.nMax = scale(_max);
        si.nMin = scale(_min);
        si.nPos = scale(_value);
        si.nPage = 0;
        SetScrollInfo(_hCtl, SB_CTL, &si, TRUE);
    }

    HWND _hCtl;
    double _min;
    double _max;
    double _step;
    double _value;
};


class Static
{
public:
    void setControl(HWND hDlg, int nIDDlgItem)
    {
        _hCtl = GetDlgItem(hDlg, nIDDlgItem);
        IF_NULL_THROW(_hCtl);
    }
    void setText(std::wstring text)
    {
        IF_FALSE_THROW(SendMessage(_hCtl, WM_SETTEXT, 0, reinterpret_cast<LPARAM>(text.c_str()))==TRUE);
    }

private:
    HWND _hCtl;
};


class Dialog
{
public:
    Dialog(HINSTANCE hInst, WORD wResourceID)
      : _hInst(hInst),
        _wResourceID(wResourceID)
    { }

    INT_PTR doModal()
    {
        INT_PTR r = DialogBoxParam(_hInst, MAKEINTRESOURCE(_wResourceID), NULL,
            mainDlgProc, reinterpret_cast<LPARAM>(this));
        IF_MINUS_ONE_THROW(r);
        return r;
    }

    virtual INT_PTR onQuit() { return TRUE; }
    virtual INT_PTR onInitDialog() { return TRUE; }

private:
    INT_PTR handleMessage(UINT uMsg, WPARAM wParam, LPARAM lParam)
    {
        switch (uMsg) {
            case WM_INITDIALOG:
                return onInitDialog();
            case WM_COMMAND:
                if (LOWORD(wParam) == IDCANCEL) {
                    EndDialog(_hDlg, IDCANCEL);
                    return TRUE;
                }
                return FALSE;
            case WM_DESTROY:
                setContext(_hDlg, NULL);
                return onQuit();
        }
        return FALSE;
    }

    static INT_PTR CALLBACK mainDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam,
        LPARAM lParam)
    {
        BEGIN_CHECKED {
            Dialog* dialog;
            if (uMsg == WM_INITDIALOG) {
                dialog = reinterpret_cast<Dialog*>(lParam);
                dialog->_hDlg = hDlg;
                setContext(hDlg, dialog);
            }
            else {
                SetLastError(0);
                LONG_PTR r = GetWindowLongPtr(hDlg, GWLP_USERDATA);
                IF_ZERO_CHECK_THROW_LAST_ERROR(r);
                dialog = reinterpret_cast<Dialog*>(r);
            }
            if (dialog != NULL)
                return dialog->handleMessage(uMsg, wParam, lParam);
        } END_CHECKED(Exception& e) {
            e.display(hDlg);
        }

        return FALSE;
    }

    static void setContext(HWND hWnd, void* p)
    {
        SetLastError(0);
        IF_ZERO_CHECK_THROW_LAST_ERROR(SetWindowLongPtr(hWnd, GWLP_USERDATA,
            reinterpret_cast<LONG_PTR>(p)));
    }

    HINSTANCE _hInst;
    WORD _wResourceID;
protected:
    HWND _hDlg;
};


class Windows
{
    friend class Window;
public:
    Windows(HINSTANCE hInst) : m_hInst(hInst)
    {
        WNDCLASS wc;
        wc.style = 0;
        wc.lpfnWndProc = wndProc;
        wc.cbClsExtra = 0;
        wc.cbWndExtra = 0;
        wc.hInstance = hInst;
        wc.hIcon = NULL;
        wc.hCursor = LoadCursor(NULL, IDC_ARROW);
        wc.hbrBackground = (HBRUSH)(COLOR_WINDOW + 1);
        wc.lpszMenuName = NULL;
        wc.lpszClassName = L"AJLib Window";
        m_classAtom = RegisterClass(&wc);
        IF_ZERO_THROW(m_classAtom);
    }

    LPCWSTR className() const
    {
        return reinterpret_cast<LPCWSTR>(static_cast<UINT_PTR>(m_classAtom));
    }

    HINSTANCE instance() const { return m_hInst; }

    ~Windows()
    {
        UnregisterClass(className(), m_hInst);
    }

    static void check()
    {
        if (_failed)
            throw _exception;
    }

private:
    static LRESULT CALLBACK wndProc(HWND hWnd, UINT uMsg, WPARAM wParam,
        LPARAM lParam);

    static void setContext(HWND hWnd, void* p)
    {
        SetLastError(0);
        IF_ZERO_CHECK_THROW_LAST_ERROR(SetWindowLongPtr(hWnd, GWLP_USERDATA,
            reinterpret_cast<LONG_PTR>(p)));
    }

    HINSTANCE m_hInst;
    ATOM m_classAtom;

    static Exception _exception;
    static bool _failed;
};

bool Windows::_failed = false;
Exception Windows::_exception;


class Window
{
    friend class Windows;
public:
    class Params
    {
        friend class Window;
    public:
        Params(Windows* windows, LPCWSTR pszName)
          : _windows(windows),
            _pszName(pszName),
            _nSize(CW_USEDEFAULT, CW_USEDEFAULT) { }
        Params(Windows* windows, LPCWSTR pszName, Vector2<int> nSize)
          : _windows(windows),
            _pszName(pszName),
            _nSize(nSize) { }
    private:
        Windows* _windows;
        LPCWSTR _pszName;
        Vector2<int> _nSize;
    };

    Window(Params p)
    {
        IF_NULL_THROW(CreateWindowEx(
            0,                        // dwExStyle
            p._windows->className(),  // lpClassName
            p._pszName,               // lpWindowName
            WS_OVERLAPPEDWINDOW,      // dwStyle
            CW_USEDEFAULT,            // x
            CW_USEDEFAULT,            // y
            p._nSize.x,               // nWidth
            p._nSize.y,               // nHeight
            NULL,                     // hWndParent
            NULL,                     // hMenu
            p._windows->instance(),   // hInstance
            this));                   // lpParam
    }

    ~Window()
    {
        DestroyWindow(_hWnd);
    }

    void show(int nShowCmd)
    {
        ShowWindow(_hWnd, nShowCmd);
    }
    void resize(Vector2<int> size)
    {
        IF_ZERO_THROW(SetWindowPos(
            _hWnd,                                // hWnd
            NULL,                                 // hWndInsertAfter
            0,                                    // X
            0,                                    // Y
            size.x,                               // cx
            size.y,                               // cy
            SWP_NOZORDER | SWP_NOMOVE |
            SWP_NOACTIVATE | SWP_NOREPOSITION));  // uFlags
    }
    Vector2<int> getSize()
    {
        RECT rect;
        IF_ZERO_THROW(GetClientRect(_hWnd, &rect));
        return Vector2<int>(rect.right, rect.bottom);
    }
    operator HWND() const { return _hWnd; }

private:
    void setHwnd(HWND hWnd) { _hWnd = hWnd; }

    virtual void onDestroy() { }

protected:
    virtual LRESULT handleMessage(UINT uMsg, WPARAM wParam, LPARAM lParam)
    {
        switch (uMsg) {
            case WM_NCDESTROY:
                onDestroy();
                Windows::setContext(_hWnd, NULL);
                break;
        }
        return DefWindowProc(_hWnd, uMsg, wParam, lParam);
    }

    HWND _hWnd;
};


template<class Base> class RootWindow : public Base
{
public:
    class Params
    {
        friend class RootWindow;
    public:
        Params(typename Base::Params bp) : _bp(bp) { }
    private:
        typename Base::Params _bp;
    };

    RootWindow(Params p) : Base(p._bp) { }

    void onDestroy()
    {
        // Death of the root window ends the thread
        PostQuitMessage(0);
    }
protected:
    virtual LRESULT handleMessage(UINT uMsg, WPARAM wParam, LPARAM lParam)
    {
        switch (uMsg) {
            case WM_KEYDOWN:
                if (wParam == VK_ESCAPE)
                    onDestroy();
        }

        return Base::handleMessage(uMsg, wParam, lParam);
    }
};


LRESULT CALLBACK Windows::wndProc(HWND hWnd, UINT uMsg, WPARAM wParam,
    LPARAM lParam)
{
    BEGIN_CHECKED {
        Window* window = 0;
        switch (uMsg) {
            case WM_NCCREATE:
                {
                    LPCREATESTRUCT lpcs =
                        reinterpret_cast<LPCREATESTRUCT>(lParam);
                    window = reinterpret_cast<Window *>(lpcs->lpCreateParams);
                    window->_hWnd = hWnd;
                    setContext(hWnd, window);
                }
                break;
            default:
                SetLastError(0);
                LONG_PTR r = GetWindowLongPtr(hWnd, GWLP_USERDATA);
                IF_ZERO_CHECK_THROW_LAST_ERROR(r);
                window = reinterpret_cast<Window *>(r);
                break;
        }
        LRESULT result;
        if (window != 0)
            result = window->handleMessage(uMsg, wParam, lParam);
        if (uMsg == WM_NCDESTROY)
            setContext(hWnd, 0);
        if (window != 0)
            return result;
    } END_CHECKED(Exception& e) {
        PostQuitMessage(0);
        _exception = e;
        _failed = true;
    }

    // The first message sent to a window is WM_GETMINMAXINFO rather than
    // WM_NCCREATE. Since we don't have a context at that point, delegate this
    // message to DefWindowProc.
    return DefWindowProc(hWnd, uMsg, wParam, lParam);
}


class DeviceContext
{
public:
    void NoFailSelectObject(HGDIOBJ hObject)
    {
        ::SelectObject(_hdc, hObject);
    }
    void SelectObject(HGDIOBJ hObject)
    {
        IF_NULL_THROW(::SelectObject(_hdc, hObject));
    }
    void SetROP2(int fnDrawMode)
    {
        IF_ZERO_THROW(::SetROP2(_hdc, fnDrawMode));
    }
    void Polyline(CONST POINT* lppt, int cPoints)
    {
        IF_ZERO_THROW(::Polyline(_hdc, lppt, cPoints));
    }
protected:
    HDC _hdc;
};


class PaintHandle : public DeviceContext
{
public:
    PaintHandle(const Window& window) : _window(window)
    {
        IF_NULL_THROW(BeginPaint(_window, &_ps));
        _hdc = _ps.hdc;
    }
    ~PaintHandle()
    {
        EndPaint(_window, &_ps);
    }
    operator HDC() const { return _ps.hdc; }
    Vector2<int> topLeft() const { return Vector2<int>(_ps.rcPaint.left, _ps.rcPaint.top); }
    Vector2<int> bottomRight() const { return Vector2<int>(_ps.rcPaint.right, _ps.rcPaint.bottom); }
    bool zeroArea() const { return (topLeft()-bottomRight()).zeroArea(); }
private:
    const Window& _window;
    PAINTSTRUCT _ps;
};


template<class Base, class ImageType> class ImageWindow : public Base
{
public:
    class Params
    {
        friend class ImageWindow;
    public:
        Params(typename Base::Params bp, ImageType* image)
          : _bp(bp),
            _image(image) { }
    private:
        typename Base::Params _bp;
        ImageType* _image;
    };

    ImageWindow(Params p) : Base(p._bp), _image(p._image) { }

    void invalidate() { IF_ZERO_THROW(InvalidateRect(_hWnd, NULL, FALSE)); }

protected:
    virtual LRESULT handleMessage(UINT uMsg, WPARAM wParam, LPARAM lParam)
    {
        switch (uMsg) {
            case WM_PAINT:
                onPaint();
                return 0;

            case WM_SIZE:
                onResize(Vector2<int>(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam)));
                return 0;
        }

        return Base::handleMessage(uMsg, wParam, lParam);
    }

    virtual bool onPaint()
    {
        return doPaint();
    }

    virtual bool doPaint()
    {
        PaintHandle paint(*this);
        if (paint.zeroArea())
            return false;
        _image->paint(paint);
        return true;
    }

protected:
    void onResize(Vector2<int> size)
    {
        _image->resize(size);
        invalidate();
    }

    ImageType* _image;
};


template<class Base> class AnimatedWindow : public Base
{
public:
    class Params
    {
        friend class AnimatedWindow;
    public:
        Params(typename Base::Params bp, int rate=20) : _bp(bp), _rate(rate) { }
    private:
        typename Base::Params _bp;
        int _rate;
    };

    AnimatedWindow(Params p)
      : Base(p._bp),
        _rate(p._rate),
        _timerExpired(true)
    { }

protected:
    virtual LRESULT handleMessage(UINT uMsg, WPARAM wParam, LPARAM lParam)
    {
        switch (uMsg) {
            case WM_TIMER:
                if (wParam == 1) {
                    invalidate();
                    KillTimer(_hWnd, _timer);
                    _timerExpired = true;
                }
                return 0;
        }

        return Base::handleMessage(uMsg, wParam, lParam);
    }

    virtual bool onPaint()
    {
        _image->draw();
        //bool painted = Base::onPaint();
        //if (painted && _timerExpired)
        //    resetTimer();
        //return painted;
        if (_timerExpired)
            resetTimer();
        return Base::onPaint();
    }

    virtual bool doPaint()
    {
        return Base::doPaint();
    }

protected:
    void resetTimer()
    {
        _timer = SetTimer(_hWnd, 1, 1000/_rate, NULL);
        IF_ZERO_THROW(_timer);
        _timerExpired = false;
    }

    UINT_PTR _timer;
    bool _timerExpired;
    int _rate;
};


class Image
{
public:
    Image() : _bits(NULL) { }
    Image(Vector2<int> size)
      : _size(size),
        _byteWidth(_size.x*4),
        _bits(_byteWidth*_size.y)
    {
        if (_size.x == 0 || _size.y == 0)
            return;

        ZeroMemory(&_bmi, sizeof(BITMAPINFO));
        _bmi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
        _bmi.bmiHeader.biWidth = _size.x;
        _bmi.bmiHeader.biHeight = -_size.y;
        _bmi.bmiHeader.biPlanes = 1;
        _bmi.bmiHeader.biBitCount = 32;
        _bmi.bmiHeader.biCompression = BI_RGB;
        _bmi.bmiHeader.biSizeImage = 0;
        _bmi.bmiHeader.biXPelsPerMeter = 0;
        _bmi.bmiHeader.biYPelsPerMeter = 0;
        _bmi.bmiHeader.biClrUsed = 0;
        _bmi.bmiHeader.biClrImportant = 0;

        draw();
    }

    void resize(Vector2<int> size)
    {
        if (size != _size) {
            Image image(size);
            swap(image);
            doResize();
            draw();
        }
    }

    void paint(const PaintHandle& paint)
    {
        if (_bits.size() == 0)
            return;
        Vector2<int> topLeft = paint.topLeft();
        Vector2<int> bottomRight = paint.bottomRight();
        Vector2<int> size = bottomRight-topLeft;
        IF_ZERO_THROW(SetDIBitsToDevice(
            paint,
            topLeft.x,
            topLeft.y,
            size.x,
            size.y,
            topLeft.x,
            (_size-bottomRight).y,
            0,
            _size.y,
            getBits(),
            &_bmi,
            DIB_RGB_COLORS));
    }

    virtual void draw() { }

private:
    BITMAPINFO _bmi;

    void swap(Image& other)
    {
        Vector2<int> size = _size; _size = other._size; other._size = size;
        int byteWidth = _byteWidth; _byteWidth = other._byteWidth; other._byteWidth = byteWidth;
        BITMAPINFO bmi = _bmi; _bmi = other._bmi; other._bmi = bmi;
        _bits.swap(other._bits);
    }

protected:
    virtual void doResize() { }

    Byte* getBits() { return &_bits[0]; }

    Vector2<int> _size;
    int _byteWidth;
private:
    std::vector<Byte> _bits;
};


class Pen
{
public:
    Pen(int fnPenStyle, int nWidth, COLORREF crColor)
    {
        m_hPen = CreatePen(fnPenStyle, nWidth, crColor);
        IF_NULL_THROW(m_hPen);
    }
    ~Pen() { DeleteObject(m_hPen); }
    operator HPEN() { return m_hPen; }
private:
    HPEN m_hPen;
};


class WindowDeviceContext : public DeviceContext
{
public:
    WindowDeviceContext(const Window& window)
      : _window(&window)
    {
        _hdc = GetDC(window);
        IF_NULL_THROW(_hdc);
    }
    ~WindowDeviceContext()
    {
        ReleaseDC(*_window, _hdc);
    }
private:
    const Window* _window;
};


class SelectedPen
{
public:
    SelectedPen(DeviceContext* dc, Pen* pen) : _dc(dc)
    {
        _dc->SelectObject(*pen);
    }
    ~SelectedPen()
    {
        _dc->NoFailSelectObject(GetStockObject(BLACK_PEN));
    }
private:
    DeviceContext* _dc;
};

#endif // AJLIB_USER_H
